# Copyright (C) 2020-2022 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# pylint: disable=too-many-lines, line-too-long, invalid-name

import unittest
from contextlib import contextmanager
from dataclasses import dataclass, field
from datetime import datetime
from pathlib import Path
from typing import Iterable, Optional, Union
from unittest.mock import AsyncMock, MagicMock, call, patch

from httpx import HTTPStatusError, Request, Response

from pontos.git.git import ConfigScope, Git
from pontos.git.status import StatusEntry
from pontos.github.actions.errors import GitHubActionsError
from pontos.release.create import (
    CreateReleaseReturnValue,
    ReleaseInformation,
    create_release,
)
from pontos.release.main import parse_args
from pontos.terminal.terminal import Terminal
from pontos.testing import temp_directory, temp_git_repository
from pontos.version import VersionError, VersionUpdate
from pontos.version.commands import GoVersionCommand
from pontos.version.schemes._pep440 import PEP440Version, PEP440VersioningScheme


def mock_terminal() -> MagicMock:
    return MagicMock(spec=Terminal)


def str_or_list(values: Union[str, Iterable[str]]) -> Iterable[str]:
    if values and isinstance(values, str):
        return [values]
    return values


@contextmanager
def setup_go_project(
    *, current_version: str, tags: Union[str, Iterable[str]] = None
) -> Path:
    with temp_git_repository() as tmp_git:
        git = Git(tmp_git)

        git.config("commit.gpgSign", "false", scope=ConfigScope.LOCAL)
        git.config("tag.gpgSign", "false", scope=ConfigScope.LOCAL)
        git.config("tag.sort", "refname", scope=ConfigScope.LOCAL)

        git.add_remote("origin", "http://foo/bar.git")

        go = GoVersionCommand(PEP440VersioningScheme)
        go.project_file_path.touch()
        update = go.update_version(new_version=PEP440Version(current_version))

        git.add(update.changed_files)
        git.add(go.project_file_path)
        git.commit("Create initial release")

        if tags:
            for tag in str_or_list(tags):
                git.tag(f"v{tag}")

        yield tmp_git


class ReleaseInformationTestCase(unittest.TestCase):
    def test_release_info(self):
        release_info = ReleaseInformation(
            last_release_version=PEP440Version.from_string("1.2.3"),
            release_version=PEP440Version.from_string("2.0.0"),
            git_release_tag="v2.0.0",
            next_version=PEP440Version.from_string("2.0.1.dev1"),
        )

        self.assertEqual(
            release_info.last_release_version,
            PEP440Version.from_string("1.2.3"),
        )
        self.assertEqual(
            release_info.release_version, PEP440Version.from_string("2.0.0")
        )
        self.assertEqual(release_info.git_release_tag, "v2.0.0")
        self.assertEqual(
            release_info.next_version, PEP440Version.from_string("2.0.1.dev1")
        )

    @patch.dict("os.environ", {}, clear=True)
    def test_no_github_output(self):
        release_info = ReleaseInformation(
            last_release_version=PEP440Version.from_string("1.2.3"),
            release_version=PEP440Version.from_string("2.0.0"),
            git_release_tag="v2.0.0",
            next_version=PEP440Version.from_string("2.0.1.dev1"),
        )

        with self.assertRaisesRegex(
            GitHubActionsError,
            "GITHUB_OUTPUT environment variable not set. Can't write "
            "action output.",
        ):
            release_info.write_github_output()

    def test_github_output(self):
        expected = """last-release-version=1.2.3
release-version=2.0.0
git-release-tag=v2.0.0
next-version=2.0.1.dev1
"""
        with temp_directory() as temp_dir:
            out_file = temp_dir / "out.txt"
            with patch.dict(
                "os.environ", {"GITHUB_OUTPUT": str(out_file.absolute())}
            ):
                release_info = ReleaseInformation(
                    last_release_version=PEP440Version.from_string("1.2.3"),
                    release_version=PEP440Version.from_string("2.0.0"),
                    git_release_tag="v2.0.0",
                    next_version=PEP440Version.from_string("2.0.1.dev1"),
                )

                release_info.write_github_output()

            self.assertTrue(out_file.exists())
            actual = out_file.read_text(encoding="utf8")
            self.assertEqual(actual, expected)


@patch.dict(
    "os.environ",
    {
        "GITHUB_TOKEN": "foo",
        "GITHUB_USER": "user",
        "GPG_SIGNING_KEY": "1234",
    },
)
class CreateReleaseTestCase(unittest.TestCase):
    @patch("pontos.release.create.Git", autospec=True)
    @patch("pontos.release.create.get_last_release_version", autospec=True)
    @patch(
        "pontos.release.create.CreateReleaseCommand._create_release",
        autospec=True,
    )
    @patch(
        "pontos.release.create.CreateReleaseCommand._create_changelog",
        autospec=True,
    )
    @patch("pontos.release.create.Project._gather_commands", autospec=True)
    def test_release_version(
        self,
        gather_commands_mock: MagicMock,
        create_changelog_mock: MagicMock,
        create_release_mock: AsyncMock,
        get_last_release_version_mock: MagicMock,
        git_mock: MagicMock,
    ):
        current_version = PEP440Version("0.0.1")
        release_version = PEP440Version("0.0.2")
        next_version = PEP440Version("1.0.0.dev1")
        command_mock = MagicMock(spec=GoVersionCommand)
        gather_commands_mock.return_value = [command_mock]
        create_changelog_mock.return_value = "A Changelog"
        get_last_release_version_mock.return_value = current_version
        command_mock.update_version.side_effect = [
            VersionUpdate(
                previous=current_version,
                new=release_version,
                changed_files=["MyProject.conf"],
            ),
            VersionUpdate(
                previous=release_version,
                new=next_version,
                changed_files=["MyProject.conf"],
            ),
        ]
        git_instance_mock: MagicMock = git_mock.return_value
        git_instance_mock.status.return_value = [
            StatusEntry("M  MyProject.conf")
        ]

        _, token, args = parse_args(
            [
                "release",
                "--project",
                "foo",
                "--release-version",
                "0.0.2",
                "--next-version",
                "1.0.0.dev1",
            ]
        )

        with temp_git_repository():
            released = create_release(
                terminal=mock_terminal(),
                error_terminal=mock_terminal(),
                args=args,
                token=token,
            )

        git_instance_mock.push.assert_has_calls(
            [
                call(follow_tags=True, remote=None),
                call(follow_tags=True, remote=None),
            ],
        )

        command_mock.update_version.assert_has_calls(
            [
                call(release_version, force=False),
                call(next_version, force=False),
            ]
        )

        self.assertEqual(
            create_release_mock.await_args.args[1:],
            (release_version, "foo", "A Changelog"),
        )

        git_instance_mock.add.assert_has_calls(
            [call("MyProject.conf"), call("MyProject.conf")]
        )
        git_instance_mock.commit.assert_has_calls(
            [
                call(
                    "Automatic release to 0.0.2",
                    verify=False,
                    gpg_signing_key="1234",
                ),
                call(
                    "Automatic adjustments after release\n\n"
                    "* Update to version 1.0.0.dev1\n",
                    verify=False,
                    gpg_signing_key="1234",
                ),
            ]
        )
        git_instance_mock.tag.assert_called_once_with(
            "v0.0.2", gpg_key_id="1234", message="Automatic release to 0.0.2"
        )

        self.assertEqual(released, CreateReleaseReturnValue.SUCCESS)

    @patch("pontos.release.create.Git", autospec=True)
    @patch("pontos.release.create.get_last_release_version", autospec=True)
    @patch(
        "pontos.release.create.CreateReleaseCommand._create_release",
        autospec=True,
    )
    @patch(
        "pontos.release.create.CreateReleaseCommand._create_changelog",
        autospec=True,
    )
    @patch("pontos.release.create.Project._gather_commands", autospec=True)
    def test_initial_release_version(
        self,
        gather_commands_mock: MagicMock,
        create_changelog_mock: MagicMock,
        create_release_mock: AsyncMock,
        get_last_release_version_mock: MagicMock,
        git_mock: MagicMock,
    ):
        current_version = PEP440Version("0.0.1")
        release_version = PEP440Version("1.0.0")
        next_version = PEP440Version("1.0.1.dev1")
        command_mock = MagicMock(spec=GoVersionCommand)
        gather_commands_mock.return_value = [command_mock]
        create_changelog_mock.return_value = "A Changelog"
        get_last_release_version_mock.return_value = None
        command_mock.update_version.side_effect = [
            VersionUpdate(
                previous=current_version,
                new=release_version,
                changed_files=["MyProject.conf"],
            ),
            VersionUpdate(
                previous=release_version,
                new=next_version,
                changed_files=["MyProject.conf"],
            ),
        ]
        git_instance_mock: MagicMock = git_mock.return_value
        git_instance_mock.status.return_value = [
            StatusEntry("M  MyProject.conf")
        ]

        _, token, args = parse_args(
            [
                "release",
                "--project",
                "foo",
                "--release-version",
                "1.0.0",
            ]
        )

        with temp_git_repository():
            released = create_release(
                terminal=mock_terminal(),
                error_terminal=mock_terminal(),
                args=args,
                token=token,
            )

        git_instance_mock.push.assert_has_calls(
            [
                call(follow_tags=True, remote=None),
                call(follow_tags=True, remote=None),
            ],
        )

        command_mock.update_version.assert_has_calls(
            [
                call(release_version, force=False),
                call(next_version, force=False),
            ]
        )

        self.assertEqual(
            create_release_mock.await_args.args[1:],
            (release_version, "foo", "A Changelog"),
        )

        git_instance_mock.add.assert_has_calls(
            [call("MyProject.conf"), call("MyProject.conf")]
        )
        git_instance_mock.commit.assert_has_calls(
            [
                call(
                    "Automatic release to 1.0.0",
                    verify=False,
                    gpg_signing_key="1234",
                ),
                call(
                    "Automatic adjustments after release\n\n"
                    "* Update to version 1.0.1.dev1\n",
                    verify=False,
                    gpg_signing_key="1234",
                ),
            ]
        )
        git_instance_mock.tag.assert_called_once_with(
            "v1.0.0", gpg_key_id="1234", message="Automatic release to 1.0.0"
        )

        self.assertEqual(released, CreateReleaseReturnValue.SUCCESS)

    @patch("pontos.release.create.Git", autospec=True)
    @patch("pontos.release.create.get_last_release_version", autospec=True)
    @patch(
        "pontos.release.create.CreateReleaseCommand._create_release",
        autospec=True,
    )
    @patch(
        "pontos.release.create.CreateReleaseCommand._create_changelog",
        autospec=True,
    )
    @patch("pontos.release.create.Project._gather_commands", autospec=True)
    def test_release_patch(
        self,
        gather_commands_mock: MagicMock,
        create_changelog_mock: MagicMock,
        create_release_mock: AsyncMock,
        get_last_release_version_mock: MagicMock,
        git_mock: MagicMock,
    ):
        current_version = PEP440Version("0.0.1")
        release_version = PEP440Version("0.0.2")
        next_version = PEP440Version("1.0.0.dev1")
        command_mock = MagicMock(spec=GoVersionCommand)
        gather_commands_mock.return_value = [command_mock]
        create_changelog_mock.return_value = "A Changelog"
        get_last_release_version_mock.return_value = current_version
        command_mock.update_version.side_effect = [
            VersionUpdate(
                previous=current_version,
                new=release_version,
                changed_files=["MyProject.conf"],
            ),
            VersionUpdate(
                previous=release_version,
                new=next_version,
                changed_files=["MyProject.conf"],
            ),
        ]
        git_instance_mock: MagicMock = git_mock.return_value
        git_instance_mock.status.return_value = [
            StatusEntry("M  MyProject.conf")
        ]

        _, token, args = parse_args(
            [
                "release",
                "--project",
                "foo",
                "--release-type",
                "patch",
                "--next-version",
                "1.0.0.dev1",
            ]
        )

        with temp_git_repository():
            released = create_release(
                terminal=mock_terminal(),
                error_terminal=mock_terminal(),
                args=args,
                token=token,
            )

        git_instance_mock.push.assert_has_calls(
            [
                call(follow_tags=True, remote=None),
                call(follow_tags=True, remote=None),
            ],
        )
        command_mock.update_version.assert_has_calls(
            [
                call(release_version, force=False),
                call(next_version, force=False),
            ],
        )

        self.assertEqual(
            create_release_mock.await_args.args[1:],
            (release_version, "foo", "A Changelog"),
        )

        git_instance_mock.add.assert_has_calls(
            [call("MyProject.conf"), call("MyProject.conf")]
        )
        git_instance_mock.commit.assert_has_calls(
            [
                call(
                    "Automatic release to 0.0.2",
                    verify=False,
                    gpg_signing_key="1234",
                ),
                call(
                    "Automatic adjustments after release\n\n"
                    "* Update to version 1.0.0.dev1\n",
                    verify=False,
                    gpg_signing_key="1234",
                ),
            ]
        )
        git_instance_mock.tag.assert_called_once_with(
            "v0.0.2", gpg_key_id="1234", message="Automatic release to 0.0.2"
        )

        self.assertEqual(released, CreateReleaseReturnValue.SUCCESS)

    @patch("pontos.release.create.Git", autospec=True)
    @patch("pontos.release.create.get_last_release_version", autospec=True)
    @patch(
        "pontos.release.create.CreateReleaseCommand._create_release",
        autospec=True,
    )
    @patch(
        "pontos.release.create.CreateReleaseCommand._create_changelog",
        autospec=True,
    )
    @patch("pontos.release.create.Project._gather_commands", autospec=True)
    def test_release_calendar(
        self,
        gather_commands_mock: MagicMock,
        create_changelog_mock: MagicMock,
        create_release_mock: AsyncMock,
        get_last_release_version_mock: MagicMock,
        git_mock: MagicMock,
    ):
        today = datetime.today()
        current_version = PEP440Version("0.0.1")
        release_version = PEP440Version(f"{today.year % 100}.{today.month}.0")
        next_version = PEP440Version(f"{today.year % 100}.{today.month}.1.dev1")
        command_mock = MagicMock(spec=GoVersionCommand)
        gather_commands_mock.return_value = [command_mock]
        create_changelog_mock.return_value = "A Changelog"
        get_last_release_version_mock.return_value = current_version
        command_mock.update_version.side_effect = [
            VersionUpdate(
                previous=current_version,
                new=release_version,
                changed_files=["MyProject.conf"],
            ),
            VersionUpdate(
                previous=release_version,
                new=next_version,
                changed_files=["MyProject.conf"],
            ),
        ]
        git_instance_mock: MagicMock = git_mock.return_value
        git_instance_mock.status.return_value = [
            StatusEntry("M  MyProject.conf")
        ]

        _, token, args = parse_args(
            [
                "release",
                "--project",
                "foo",
                "--release-type",
                "calendar",
            ]
        )

        with temp_git_repository():
            released = create_release(
                terminal=mock_terminal(),
                error_terminal=mock_terminal(),
                args=args,
                token=token,
            )

        git_instance_mock.push.assert_has_calls(
            [
                call(follow_tags=True, remote=None),
                call(follow_tags=True, remote=None),
            ],
        )

        command_mock.update_version.assert_has_calls(
            [
                call(release_version, force=False),
                call(next_version, force=False),
            ]
        )

        self.assertEqual(
            create_release_mock.await_args.args[1:],
            (release_version, "foo", "A Changelog"),
        )

        git_instance_mock.add.assert_has_calls(
            [call("MyProject.conf"), call("MyProject.conf")]
        )
        git_instance_mock.commit.assert_has_calls(
            [
                call(
                    f"Automatic release to {release_version}",
                    verify=False,
                    gpg_signing_key="1234",
                ),
                call(
                    "Automatic adjustments after release\n\n"
                    f"* Update to version {next_version}\n",
                    verify=False,
                    gpg_signing_key="1234",
                ),
            ]
        )
        git_instance_mock.tag.assert_called_once_with(
            f"v{release_version}",
            gpg_key_id="1234",
            message=f"Automatic release to {release_version}",
        )

        self.assertEqual(released, CreateReleaseReturnValue.SUCCESS)

    @patch("pontos.release.create.Git", autospec=True)
    @patch("pontos.release.create.get_last_release_version", autospec=True)
    @patch(
        "pontos.release.create.CreateReleaseCommand._create_release",
        autospec=True,
    )
    @patch(
        "pontos.release.create.CreateReleaseCommand._create_changelog",
        autospec=True,
    )
    @patch("pontos.release.create.Project._gather_commands", autospec=True)
    def test_release_minor(
        self,
        gather_commands_mock: MagicMock,
        create_changelog_mock: MagicMock,
        create_release_mock: AsyncMock,
        get_last_release_version_mock: MagicMock,
        git_mock: MagicMock,
    ):
        current_version = PEP440Version("0.0.1")
        release_version = PEP440Version("0.1.0")
        next_version = PEP440Version("0.1.1.dev1")
        command_mock = MagicMock(spec=GoVersionCommand)
        gather_commands_mock.return_value = [command_mock]
        create_changelog_mock.return_value = "A Changelog"
        get_last_release_version_mock.return_value = current_version
        command_mock.update_version.side_effect = [
            VersionUpdate(
                previous=current_version,
                new=release_version,
                changed_files=["MyProject.conf"],
            ),
            VersionUpdate(
                previous=release_version,
                new=next_version,
                changed_files=["MyProject.conf"],
            ),
        ]
        git_instance_mock: MagicMock = git_mock.return_value
        git_instance_mock.status.return_value = [
            StatusEntry("M  MyProject.conf")
        ]

        _, token, args = parse_args(
            [
                "release",
                "--project",
                "foo",
                "--release-type",
                "minor",
                "--next-version",
                "0.1.1.dev1",
            ]
        )

        with temp_git_repository():
            released = create_release(
                terminal=mock_terminal(),
                error_terminal=mock_terminal(),
                args=args,
                token=token,
            )

        git_instance_mock.push.assert_has_calls(
            [
                call(follow_tags=True, remote=None),
                call(follow_tags=True, remote=None),
            ],
        )
        command_mock.update_version.assert_has_calls(
            [
                call(release_version, force=False),
                call(next_version, force=False),
            ],
        )

        self.assertEqual(
            create_release_mock.await_args.args[1:],
            (release_version, "foo", "A Changelog"),
        )

        git_instance_mock.add.assert_has_calls(
            [call("MyProject.conf"), call("MyProject.conf")]
        )
        git_instance_mock.commit.assert_has_calls(
            [
                call(
                    "Automatic release to 0.1.0",
                    verify=False,
                    gpg_signing_key="1234",
                ),
                call(
                    "Automatic adjustments after release\n\n"
                    "* Update to version 0.1.1.dev1\n",
                    verify=False,
                    gpg_signing_key="1234",
                ),
            ]
        )
        git_instance_mock.tag.assert_called_once_with(
            "v0.1.0", gpg_key_id="1234", message="Automatic release to 0.1.0"
        )

        self.assertEqual(released, CreateReleaseReturnValue.SUCCESS)

    @patch("pontos.release.create.Git", autospec=True)
    @patch("pontos.release.create.get_last_release_version", autospec=True)
    @patch(
        "pontos.release.create.CreateReleaseCommand._create_release",
        autospec=True,
    )
    @patch(
        "pontos.release.create.CreateReleaseCommand._create_changelog",
        autospec=True,
    )
    @patch("pontos.release.create.Project._gather_commands", autospec=True)
    def test_release_major(
        self,
        gather_commands_mock: MagicMock,
        create_changelog_mock: MagicMock,
        create_release_mock: AsyncMock,
        get_last_release_version_mock: MagicMock,
        git_mock: MagicMock,
    ):
        current_version = PEP440Version("0.0.1")
        release_version = PEP440Version("1.0.0")
        next_version = PEP440Version("1.0.1.dev1")
        command_mock = MagicMock(spec=GoVersionCommand)
        gather_commands_mock.return_value = [command_mock]
        create_changelog_mock.return_value = "A Changelog"
        get_last_release_version_mock.return_value = current_version
        command_mock.update_version.side_effect = [
            VersionUpdate(
                previous=current_version,
                new=release_version,
                changed_files=["MyProject.conf"],
            ),
            VersionUpdate(
                previous=release_version,
                new=next_version,
                changed_files=["MyProject.conf"],
            ),
        ]
        git_instance_mock: MagicMock = git_mock.return_value
        git_instance_mock.status.return_value = [
            StatusEntry("M  MyProject.conf")
        ]

        _, token, args = parse_args(
            [
                "release",
                "--project",
                "foo",
                "--release-type",
                "major",
                "--next-version",
                "1.0.1.dev1",
            ]
        )

        with temp_git_repository():
            released = create_release(
                terminal=mock_terminal(),
                error_terminal=mock_terminal(),
                args=args,
                token=token,
            )

        git_instance_mock.push.assert_has_calls(
            [
                call(follow_tags=True, remote=None),
                call(follow_tags=True, remote=None),
            ],
        )
        command_mock.update_version.assert_has_calls(
            [
                call(release_version, force=False),
                call(next_version, force=False),
            ],
        )

        self.assertEqual(
            create_release_mock.await_args.args[1:],
            (release_version, "foo", "A Changelog"),
        )

        git_instance_mock.add.assert_has_calls(
            [call("MyProject.conf"), call("MyProject.conf")]
        )
        git_instance_mock.commit.assert_has_calls(
            [
                call(
                    "Automatic release to 1.0.0",
                    verify=False,
                    gpg_signing_key="1234",
                ),
                call(
                    "Automatic adjustments after release\n\n"
                    "* Update to version 1.0.1.dev1\n",
                    verify=False,
                    gpg_signing_key="1234",
                ),
            ]
        )
        git_instance_mock.tag.assert_called_once_with(
            "v1.0.0", gpg_key_id="1234", message="Automatic release to 1.0.0"
        )

        self.assertEqual(released, CreateReleaseReturnValue.SUCCESS)

    @patch("pontos.release.create.Git", autospec=True)
    @patch("pontos.release.create.get_last_release_version", autospec=True)
    @patch(
        "pontos.release.create.CreateReleaseCommand._create_release",
        autospec=True,
    )
    @patch(
        "pontos.release.create.CreateReleaseCommand._create_changelog",
        autospec=True,
    )
    @patch("pontos.release.create.Project._gather_commands", autospec=True)
    def test_release_alpha(
        self,
        gather_commands_mock: MagicMock,
        create_changelog_mock: MagicMock,
        create_release_mock: AsyncMock,
        get_last_release_version_mock: MagicMock,
        git_mock: MagicMock,
    ):
        current_version = PEP440Version("0.0.1")
        release_version = PEP440Version("0.0.2a1")
        next_version = PEP440Version("0.0.2a2.dev1")
        command_mock = MagicMock(spec=GoVersionCommand)
        gather_commands_mock.return_value = [command_mock]
        create_changelog_mock.return_value = "A Changelog"
        get_last_release_version_mock.return_value = current_version
        command_mock.update_version.side_effect = [
            VersionUpdate(
                previous=current_version,
                new=release_version,
                changed_files=["MyProject.conf"],
            ),
            VersionUpdate(
                previous=release_version,
                new=next_version,
                changed_files=["MyProject.conf"],
            ),
        ]
        git_instance_mock: MagicMock = git_mock.return_value
        git_instance_mock.status.return_value = [
            StatusEntry("M  MyProject.conf")
        ]

        _, token, args = parse_args(
            [
                "release",
                "--project",
                "foo",
                "--release-type",
                "alpha",
            ]
        )

        with temp_git_repository():
            released = create_release(
                terminal=mock_terminal(),
                error_terminal=mock_terminal(),
                args=args,
                token=token,
            )

        git_instance_mock.push.assert_has_calls(
            [
                call(follow_tags=True, remote=None),
                call(follow_tags=True, remote=None),
            ],
        )
        command_mock.update_version.assert_has_calls(
            [
                call(release_version, force=False),
                call(next_version, force=False),
            ],
        )

        self.assertEqual(
            create_release_mock.await_args.args[1:],
            (release_version, "foo", "A Changelog"),
        )

        git_instance_mock.add.assert_has_calls(
            [call("MyProject.conf"), call("MyProject.conf")]
        )
        git_instance_mock.commit.assert_has_calls(
            [
                call(
                    "Automatic release to 0.0.2a1",
                    verify=False,
                    gpg_signing_key="1234",
                ),
                call(
                    "Automatic adjustments after release\n\n"
                    "* Update to version 0.0.2a2.dev1\n",
                    verify=False,
                    gpg_signing_key="1234",
                ),
            ]
        )
        git_instance_mock.tag.assert_called_once_with(
            "v0.0.2a1",
            gpg_key_id="1234",
            message="Automatic release to 0.0.2a1",
        )

        self.assertEqual(released, CreateReleaseReturnValue.SUCCESS)

    @patch("pontos.release.create.Git", autospec=True)
    @patch("pontos.release.create.get_last_release_version", autospec=True)
    @patch(
        "pontos.release.create.CreateReleaseCommand._create_release",
        autospec=True,
    )
    @patch(
        "pontos.release.create.CreateReleaseCommand._create_changelog",
        autospec=True,
    )
    @patch("pontos.release.create.Project._gather_commands", autospec=True)
    def test_release_beta(
        self,
        gather_commands_mock: MagicMock,
        create_changelog_mock: MagicMock,
        create_release_mock: AsyncMock,
        get_last_release_version_mock: MagicMock,
        git_mock: MagicMock,
    ):
        current_version = PEP440Version("0.0.1")
        release_version = PEP440Version("0.0.2b1")
        next_version = PEP440Version("0.0.2b2.dev1")
        command_mock = MagicMock(spec=GoVersionCommand)
        gather_commands_mock.return_value = [command_mock]
        create_changelog_mock.return_value = "A Changelog"
        get_last_release_version_mock.return_value = current_version
        command_mock.update_version.side_effect = [
            VersionUpdate(
                previous=current_version,
                new=release_version,
                changed_files=["MyProject.conf"],
            ),
            VersionUpdate(
                previous=release_version,
                new=next_version,
                changed_files=["MyProject.conf"],
            ),
        ]
        git_instance_mock: MagicMock = git_mock.return_value
        git_instance_mock.status.return_value = [
            StatusEntry("M  MyProject.conf")
        ]

        _, token, args = parse_args(
            [
                "release",
                "--project",
                "foo",
                "--release-type",
                "beta",
            ]
        )

        with temp_git_repository():
            released = create_release(
                terminal=mock_terminal(),
                error_terminal=mock_terminal(),
                args=args,
                token=token,
            )

        git_instance_mock.push.assert_has_calls(
            [
                call(follow_tags=True, remote=None),
                call(follow_tags=True, remote=None),
            ],
        )
        command_mock.update_version.assert_has_calls(
            [
                call(release_version, force=False),
                call(next_version, force=False),
            ],
        )

        self.assertEqual(
            create_release_mock.await_args.args[1:],
            (release_version, "foo", "A Changelog"),
        )

        git_instance_mock.add.assert_has_calls(
            [call("MyProject.conf"), call("MyProject.conf")]
        )
        git_instance_mock.commit.assert_has_calls(
            [
                call(
                    "Automatic release to 0.0.2b1",
                    verify=False,
                    gpg_signing_key="1234",
                ),
                call(
                    "Automatic adjustments after release\n\n"
                    "* Update to version 0.0.2b2.dev1\n",
                    verify=False,
                    gpg_signing_key="1234",
                ),
            ]
        )
        git_instance_mock.tag.assert_called_once_with(
            "v0.0.2b1",
            gpg_key_id="1234",
            message="Automatic release to 0.0.2b1",
        )

        self.assertEqual(released, CreateReleaseReturnValue.SUCCESS)

    @patch("pontos.release.create.Git", autospec=True)
    @patch("pontos.release.create.get_last_release_version", autospec=True)
    @patch(
        "pontos.release.create.CreateReleaseCommand._create_release",
        autospec=True,
    )
    @patch(
        "pontos.release.create.CreateReleaseCommand._create_changelog",
        autospec=True,
    )
    @patch("pontos.release.create.Project._gather_commands", autospec=True)
    def test_release_release_candidate(
        self,
        gather_commands_mock: MagicMock,
        create_changelog_mock: MagicMock,
        create_release_mock: AsyncMock,
        get_last_release_version_mock: MagicMock,
        git_mock: MagicMock,
    ):
        current_version = PEP440Version("0.0.1")
        release_version = PEP440Version("0.0.2rc1")
        next_version = PEP440Version("0.0.2rc2.dev1")
        command_mock = MagicMock(spec=GoVersionCommand)
        gather_commands_mock.return_value = [command_mock]
        create_changelog_mock.return_value = "A Changelog"
        get_last_release_version_mock.return_value = current_version
        command_mock.update_version.side_effect = [
            VersionUpdate(
                previous=current_version,
                new=release_version,
                changed_files=["MyProject.conf"],
            ),
            VersionUpdate(
                previous=release_version,
                new=next_version,
                changed_files=["MyProject.conf"],
            ),
        ]
        git_instance_mock: MagicMock = git_mock.return_value
        git_instance_mock.status.return_value = [
            StatusEntry("M  MyProject.conf")
        ]

        _, token, args = parse_args(
            [
                "release",
                "--project",
                "foo",
                "--release-type",
                "release-candidate",
            ]
        )

        with temp_git_repository():
            released = create_release(
                terminal=mock_terminal(),
                error_terminal=mock_terminal(),
                args=args,
                token=token,
            )

        git_instance_mock.push.assert_has_calls(
            [
                call(follow_tags=True, remote=None),
                call(follow_tags=True, remote=None),
            ],
        )
        command_mock.update_version.assert_has_calls(
            [
                call(release_version, force=False),
                call(next_version, force=False),
            ],
        )

        self.assertEqual(
            create_release_mock.await_args.args[1:],
            (release_version, "foo", "A Changelog"),
        )

        git_instance_mock.add.assert_has_calls(
            [call("MyProject.conf"), call("MyProject.conf")]
        )
        git_instance_mock.commit.assert_has_calls(
            [
                call(
                    "Automatic release to 0.0.2rc1",
                    verify=False,
                    gpg_signing_key="1234",
                ),
                call(
                    "Automatic adjustments after release\n\n"
                    "* Update to version 0.0.2rc2.dev1\n",
                    verify=False,
                    gpg_signing_key="1234",
                ),
            ]
        )
        git_instance_mock.tag.assert_called_once_with(
            "v0.0.2rc1",
            gpg_key_id="1234",
            message="Automatic release to 0.0.2rc1",
        )

        self.assertEqual(released, CreateReleaseReturnValue.SUCCESS)

    def test_no_token(
        self,
    ):
        _, _, args = parse_args(
            [
                "release",
                "--project",
                "foo",
                "--release-version",
                "0.0.1",
            ]
        )

        released = create_release(
            terminal=mock_terminal(),
            error_terminal=mock_terminal(),
            args=args,
            token=None,
        )

        self.assertEqual(released, CreateReleaseReturnValue.TOKEN_MISSING)

    @patch("pontos.release.create.get_last_release_version", autospec=True)
    def test_no_project_settings(
        self,
        get_last_release_version_mock: MagicMock,
    ):
        current_version = PEP440Version("0.0.1")
        get_last_release_version_mock.return_value = current_version

        _, token, args = parse_args(
            [
                "release",
                "--project",
                "foo",
                "--release-version",
                "0.0.1",
            ]
        )
        with temp_git_repository():
            released = create_release(
                terminal=mock_terminal(),
                error_terminal=mock_terminal(),
                args=args,
                token=token,
            )

        self.assertEqual(
            released, CreateReleaseReturnValue.PROJECT_SETTINGS_NOT_FOUND
        )

    @patch("pontos.release.create.Git", autospec=True)
    @patch("pontos.release.create.get_last_release_version", autospec=True)
    @patch(
        "pontos.release.create.CreateReleaseCommand._create_release",
        autospec=True,
    )
    @patch(
        "pontos.release.create.CreateReleaseCommand._create_changelog",
        autospec=True,
    )
    @patch("pontos.release.create.Project", autospec=True)
    def test_no_update_project(
        self,
        project_mock: MagicMock,
        create_changelog_mock: MagicMock,
        create_release_mock: AsyncMock,
        get_last_release_version_mock: MagicMock,
        git_mock: MagicMock,
    ):
        current_version = PEP440Version("0.0.1")
        release_version = PEP440Version("0.0.2")
        create_changelog_mock.return_value = "A Changelog"
        get_last_release_version_mock.return_value = current_version
        git_instance_mock: MagicMock = git_mock.return_value
        git_instance_mock.status.return_value = []

        _, token, args = parse_args(
            [
                "release",
                "--project",
                "foo",
                "--release-version",
                "0.0.2",
                "--next-version",
                "1.0.0.dev1",
                "--no-update-project",
            ]
        )

        with temp_git_repository():
            released = create_release(
                terminal=mock_terminal(),
                error_terminal=mock_terminal(),
                args=args,
                token=token,
            )

        git_instance_mock.push.assert_has_calls(
            [
                call(follow_tags=True, remote=None),
                call(follow_tags=True, remote=None),
            ],
        )

        self.assertEqual(
            create_release_mock.await_args.args[1:],
            (release_version, "foo", "A Changelog"),
        )

        git_instance_mock.add.assert_not_called()
        git_instance_mock.commit.assert_not_called()
        git_instance_mock.tag.assert_called_once_with(
            "v0.0.2", gpg_key_id="1234", message="Automatic release to 0.0.2"
        )

        self.assertEqual(released, CreateReleaseReturnValue.SUCCESS)

        project_mock.assert_not_called()

    def test_last_release_version_error(self):
        _, token, args = parse_args(
            [
                "release",
                "--project",
                "foo",
                "--release-type",
                "major",
            ]
        )
        with temp_git_repository():
            released = create_release(
                terminal=mock_terminal(),
                error_terminal=mock_terminal(),
                args=args,
                token=token,
            )

        self.assertEqual(
            released, CreateReleaseReturnValue.NO_LAST_RELEASE_VERSION
        )

    @patch("pontos.release.create.get_last_release_version", autospec=True)
    def test_no_last_release_version(
        self,
        get_last_release_version_mock: MagicMock,
    ):
        get_last_release_version_mock.return_value = None
        _, token, args = parse_args(
            [
                "release",
                "--project",
                "foo",
                "--release-type",
                "minor",
            ]
        )
        with temp_git_repository():
            released = create_release(
                terminal=mock_terminal(),
                error_terminal=mock_terminal(),
                args=args,
                token=token,
            )

        self.assertEqual(
            released, CreateReleaseReturnValue.NO_LAST_RELEASE_VERSION
        )

    @patch("pontos.release.create.get_last_release_version", autospec=True)
    @patch(
        "pontos.release.create.get_next_release_version",
        autospec=True,
    )
    def test_no_release_error(
        self,
        get_next_release_version_mock: MagicMock,
        get_last_release_version_mock: MagicMock,
    ):
        current_version = PEP440Version("0.0.1")
        get_next_release_version_mock.side_effect = VersionError("An error")
        get_last_release_version_mock.return_value = current_version

        _, token, args = parse_args(
            [
                "release",
                "--project",
                "foo",
                "--git-signing-key",
                "123",
                "--release-type",
                "patch",
            ]
        )

        with temp_git_repository():
            released = create_release(
                terminal=mock_terminal(),
                error_terminal=mock_terminal(),
                args=args,
                token=token,
            )

        self.assertEqual(released, CreateReleaseReturnValue.NO_RELEASE_VERSION)

    @patch("pontos.release.create.Git", autospec=True)
    @patch("pontos.version.helper.Git", autospec=True)
    def test_has_tag(
        self,
        git_mock1: MagicMock,
        git_mock2: MagicMock,
    ):
        tags = ["v1.0.0", "v1.0.1"]
        git_mock1.return_value.list_tags.return_value = tags
        git_mock2.return_value.list_tags.return_value = tags

        _, token, args = parse_args(
            [
                "release",
                "--release-version",
                "1.0.1",
                "--project",
                "bla",
            ]
        )

        with temp_git_repository():
            released = create_release(
                terminal=mock_terminal(),
                error_terminal=mock_terminal(),
                token=token,
                args=args,
            )

            self.assertEqual(released, CreateReleaseReturnValue.ALREADY_TAKEN)

        git_mock1.return_value.list_tags.assert_called_once()
        git_mock2.return_value.list_tags.assert_called_once()

    @patch("pontos.release.create.Git", autospec=True)
    @patch("pontos.release.create.get_last_release_version", autospec=True)
    @patch(
        "pontos.release.create.CreateReleaseCommand._create_release",
        autospec=True,
    )
    @patch("pontos.release.create.Project._gather_commands", autospec=True)
    def test_update_version_error(
        self,
        gather_commands_mock: MagicMock,
        create_release_mock: AsyncMock,
        get_last_release_version_mock: MagicMock,
        git_mock: MagicMock,
    ):
        command_mock = MagicMock(spec=GoVersionCommand)
        gather_commands_mock.return_value = [command_mock]
        command_mock.update_version.side_effect = VersionError("An error")
        current_version = PEP440Version("0.0.1rc1")
        get_last_release_version_mock.return_value = current_version

        _, token, args = parse_args(
            [
                "release",
                "--project",
                "foo",
                "--release-version",
                "0.0.1",
                "--next-version",
                "0.0.2.dev1",
            ]
        )

        with temp_git_repository():
            released = create_release(
                terminal=mock_terminal(),
                error_terminal=mock_terminal(),
                args=args,
                token=token,
            )

        git_mock.return_value.push.assert_not_called()

        command_mock.update_version.assert_called_once_with(
            PEP440Version("0.0.1"), force=False
        )

        create_release_mock.assert_not_awaited()

        git_mock.return_value.add.assert_not_called()
        git_mock.return_value.commit.assert_not_called()
        git_mock.return_value.tag.assert_not_called()

        self.assertEqual(
            released, CreateReleaseReturnValue.UPDATE_VERSION_ERROR
        )

    @patch("pontos.release.create.Git", autospec=True)
    @patch("pontos.release.create.get_last_release_version", autospec=True)
    @patch(
        "pontos.release.create.CreateReleaseCommand._create_release",
        autospec=True,
    )
    @patch(
        "pontos.release.create.CreateReleaseCommand._create_changelog",
        autospec=True,
    )
    @patch("pontos.release.create.Project._gather_commands", autospec=True)
    def test_github_create_release_failure(
        self,
        gather_commands_mock: MagicMock,
        create_changelog_mock: MagicMock,
        create_release_mock: AsyncMock,
        get_last_release_version_mock: MagicMock,
        git_mock: MagicMock,
    ):
        current_version = PEP440Version("0.0.0")
        release_version = PEP440Version("0.0.1")
        next_version = PEP440Version("0.0.2.dev1")
        command_mock = MagicMock(spec=GoVersionCommand)
        gather_commands_mock.return_value = [command_mock]
        get_last_release_version_mock.return_value = current_version
        command_mock.update_version.side_effect = [
            VersionUpdate(
                previous=current_version,
                new=release_version,
                changed_files=["MyProject.conf"],
            ),
            VersionUpdate(
                previous=release_version,
                new=next_version,
                changed_files=["MyProject.conf"],
            ),
        ]
        create_changelog_mock.return_value = "A Changelog"
        create_release_mock.side_effect = HTTPStatusError(
            "Error during a request",
            request=MagicMock(spec=Request),
            response=MagicMock(spec=Response),
        )
        git_instance_mock: MagicMock = git_mock.return_value
        git_instance_mock.status.return_value = [
            StatusEntry("M  MyProject.conf")
        ]

        _, token, args = parse_args(
            [
                "release",
                "--project",
                "foo",
                "--release-version",
                "0.0.1",
                "--next-version",
                "0.0.2.dev1",
            ]
        )

        with temp_git_repository():
            released = create_release(
                terminal=mock_terminal(),
                error_terminal=mock_terminal(),
                args=args,
                token=token,
            )

        self.assertEqual(
            released, CreateReleaseReturnValue.CREATE_RELEASE_ERROR
        )

        git_instance_mock.push.assert_called_once_with(
            follow_tags=True, remote=None
        )
        git_instance_mock.add.assert_called_once_with("MyProject.conf")
        git_instance_mock.commit.assert_called_once_with(
            "Automatic release to 0.0.1", verify=False, gpg_signing_key="1234"
        )
        git_instance_mock.tag.assert_called_once_with(
            "v0.0.1", gpg_key_id="1234", message="Automatic release to 0.0.1"
        )

        self.assertEqual(
            create_release_mock.await_args.args[1:],
            (release_version, "foo", "A Changelog"),
        )

    @patch("pontos.release.create.Git", autospec=True)
    @patch("pontos.release.create.get_last_release_version", autospec=True)
    @patch(
        "pontos.release.create.CreateReleaseCommand._create_release",
        autospec=True,
    )
    @patch(
        "pontos.release.create.CreateReleaseCommand._create_changelog",
        autospec=True,
    )
    @patch("pontos.release.create.Project._gather_commands", autospec=True)
    def test_update_version_after_release_error(
        self,
        gather_commands_mock: MagicMock,
        create_changelog_mock: MagicMock,
        create_release_mock: AsyncMock,
        get_last_release_version_mock: MagicMock,
        git_mock: MagicMock,
    ):
        current_version = PEP440Version("0.0.0")
        release_version = PEP440Version("0.0.1")
        next_version = PEP440Version("0.0.2.dev1")
        command_mock = MagicMock(spec=GoVersionCommand)
        gather_commands_mock.return_value = [command_mock]
        create_changelog_mock.return_value = "A Changelog"
        get_last_release_version_mock.return_value = current_version
        command_mock.update_version.side_effect = [
            VersionUpdate(
                previous=current_version,
                new=release_version,
                changed_files=["MyProject.conf"],
            ),
            VersionError("An error"),
        ]
        git_instance_mock: MagicMock = git_mock.return_value
        git_instance_mock.status.return_value = [
            StatusEntry("M  MyProject.conf")
        ]

        _, token, args = parse_args(
            [
                "release",
                "--project",
                "foo",
                "--release-version",
                "0.0.1",
                "--next-version",
                "0.0.2.dev1",
            ]
        )

        with temp_git_repository():
            released = create_release(
                terminal=mock_terminal(),
                error_terminal=mock_terminal(),
                args=args,
                token=token,
            )

        git_instance_mock.push.assert_called_once_with(
            follow_tags=True, remote=None
        )

        command_mock.update_version.assert_has_calls(
            [
                call(release_version, force=False),
                call(next_version, force=False),
            ]
        )

        self.assertEqual(
            create_release_mock.await_args.args[1:],
            (release_version, "foo", "A Changelog"),
        )

        git_instance_mock.add.assert_called_once_with("MyProject.conf")
        git_instance_mock.commit.assert_called_once_with(
            "Automatic release to 0.0.1", verify=False, gpg_signing_key="1234"
        )
        git_instance_mock.tag.assert_called_once_with(
            "v0.0.1", gpg_key_id="1234", message="Automatic release to 0.0.1"
        )

        self.assertEqual(
            released,
            CreateReleaseReturnValue.UPDATE_VERSION_AFTER_RELEASE_ERROR,
        )

    @patch("pontos.release.create.Git", autospec=True)
    @patch("pontos.release.create.get_last_release_version", autospec=True)
    @patch(
        "pontos.release.create.CreateReleaseCommand._create_release",
        autospec=True,
    )
    @patch(
        "pontos.release.create.CreateReleaseCommand._create_changelog",
        autospec=True,
    )
    @patch("pontos.release.create.Project._gather_commands", autospec=True)
    def test_release_to_specific_git_remote(
        self,
        gather_commands_mock: MagicMock,
        create_changelog_mock: MagicMock,
        create_release_mock: AsyncMock,
        get_last_release_version_mock: MagicMock,
        git_mock,
    ):
        current_version = PEP440Version("0.0.0")
        release_version = PEP440Version("0.0.1")
        next_version = PEP440Version("0.0.2.dev1")
        command_mock = MagicMock(spec=GoVersionCommand)
        gather_commands_mock.return_value = [command_mock]
        get_last_release_version_mock.return_value = current_version
        command_mock.update_version.side_effect = [
            VersionUpdate(
                previous=current_version,
                new=release_version,
                changed_files=["MyProject.conf"],
            ),
            VersionUpdate(
                previous=release_version,
                new=next_version,
                changed_files=["MyProject.conf"],
            ),
        ]
        create_changelog_mock.return_value = "A Changelog"
        git_instance_mock: MagicMock = git_mock.return_value
        git_instance_mock.status.return_value = [
            StatusEntry("M  MyProject.conf")
        ]

        _, token, args = parse_args(
            [
                "release",
                "--project",
                "foo",
                "--release-version",
                "0.0.1",
                "--next-version",
                "0.0.2.dev1",
                "--git-remote-name",
                "upstream",
                "--git-signing-key",
                "1234",
            ]
        )

        with temp_git_repository():
            released = create_release(
                terminal=mock_terminal(),
                error_terminal=mock_terminal(),
                args=args,
                token=token,
            )

        git_instance_mock.push.assert_has_calls(
            [
                call(follow_tags=True, remote="upstream"),
                call(follow_tags=True, remote="upstream"),
            ],
        )

        self.assertEqual(
            create_release_mock.await_args.args[1:],
            (release_version, "foo", "A Changelog"),
        )

        command_mock.update_version.assert_has_calls(
            [
                call(release_version, force=False),
                call(next_version, force=False),
            ],
        )

        git_instance_mock.add.assert_called_with("MyProject.conf")
        git_instance_mock.commit.assert_called_with(
            "Automatic adjustments after release\n\n"
            "* Update to version 0.0.2.dev1\n",
            verify=False,
            gpg_signing_key="1234",
        )
        git_instance_mock.tag.assert_called_once_with(
            "v0.0.1", gpg_key_id="1234", message="Automatic release to 0.0.1"
        )

        self.assertEqual(released, CreateReleaseReturnValue.SUCCESS)

    @patch("pontos.release.create.Git", autospec=True)
    @patch("pontos.release.create.get_last_release_version", autospec=True)
    @patch(
        "pontos.release.create.CreateReleaseCommand._create_release",
        autospec=True,
    )
    @patch(
        "pontos.release.create.CreateReleaseCommand._create_changelog",
        autospec=True,
    )
    @patch("pontos.release.create.Project._gather_commands", autospec=True)
    def test_release_without_git_prefix(
        self,
        gather_commands_mock: MagicMock,
        create_changelog_mock: MagicMock,
        create_release_mock: AsyncMock,
        get_last_release_version_mock: MagicMock,
        git_mock,
    ):
        current_version = PEP440Version("0.0.0")
        release_version = PEP440Version("0.0.1")
        next_version = PEP440Version("0.0.2.dev1")
        command_mock = MagicMock(spec=GoVersionCommand)
        gather_commands_mock.return_value = [command_mock]
        get_last_release_version_mock.return_value = current_version
        command_mock.update_version.side_effect = [
            VersionUpdate(
                previous=current_version,
                new=release_version,
                changed_files=["MyProject.conf"],
            ),
            VersionUpdate(
                previous=release_version,
                new=next_version,
                changed_files=["MyProject.conf"],
            ),
        ]
        create_changelog_mock.return_value = "A Changelog"
        git_instance_mock: MagicMock = git_mock.return_value
        git_instance_mock.status.return_value = [
            StatusEntry("M  MyProject.conf")
        ]

        _, token, args = parse_args(
            [
                "release",
                "--project",
                "foo",
                "--release-version",
                "0.0.1",
                "--next-version",
                "0.0.2.dev1",
                "--git-tag-prefix",
                "",
            ]
        )

        with temp_git_repository():
            released = create_release(
                terminal=mock_terminal(),
                error_terminal=mock_terminal(),
                args=args,
                token=token,
            )

        git_instance_mock.push.assert_has_calls(
            [
                call(follow_tags=True, remote=None),
                call(follow_tags=True, remote=None),
            ],
        )

        self.assertEqual(
            create_release_mock.await_args.args[1:],
            (release_version, "foo", "A Changelog"),
        )

        command_mock.update_version.assert_has_calls(
            [
                call(release_version, force=False),
                call(next_version, force=False),
            ],
        )

        git_instance_mock.add.assert_called_with("MyProject.conf")
        git_instance_mock.commit.assert_called_with(
            "Automatic adjustments after release\n\n"
            "* Update to version 0.0.2.dev1\n",
            verify=False,
            gpg_signing_key="1234",
        )
        git_instance_mock.tag.assert_called_once_with(
            "0.0.1", gpg_key_id="1234", message="Automatic release to 0.0.1"
        )

        self.assertEqual(released, CreateReleaseReturnValue.SUCCESS)

    @patch("pontos.release.create.Git", autospec=True)
    @patch("pontos.release.create.get_last_release_version", autospec=True)
    @patch("pontos.release.create.GitHubAsyncRESTApi", autospec=True)
    @patch(
        "pontos.release.create.CreateReleaseCommand._create_changelog",
        autospec=True,
    )
    @patch("pontos.release.create.Project._gather_commands", autospec=True)
    def test_release_github_api(
        self,
        gather_commands_mock: MagicMock,
        create_changelog_mock: MagicMock,
        github_api_mock: AsyncMock,
        get_last_release_version_mock: MagicMock,
        git_mock,
    ):
        current_version = PEP440Version("0.0.0")
        release_version = PEP440Version("0.0.1")
        next_version = PEP440Version("0.0.2.dev1")
        command_mock = MagicMock(spec=GoVersionCommand)
        gather_commands_mock.return_value = [command_mock]
        get_last_release_version_mock.return_value = current_version
        command_mock.update_version.side_effect = [
            VersionUpdate(
                previous=current_version,
                new=release_version,
                changed_files=["MyProject.conf"],
            ),
            VersionUpdate(
                previous=release_version,
                new=next_version,
                changed_files=["MyProject.conf"],
            ),
        ]
        create_changelog_mock.return_value = "A Changelog"
        create_api_mock = AsyncMock()
        github_api_mock.return_value.releases.create = create_api_mock
        git_instance_mock: MagicMock = git_mock.return_value
        git_instance_mock.status.return_value = [
            StatusEntry("M  MyProject.conf")
        ]

        _, token, args = parse_args(
            [
                "release",
                "--space",
                "bar",
                "--project",
                "foo",
                "--release-version",
                "0.0.1",
                "--next-version",
                "0.0.2.dev1",
                "--git-remote-name",
                "upstream",
            ]
        )

        with temp_git_repository():
            released = create_release(
                terminal=mock_terminal(),
                error_terminal=mock_terminal(),
                args=args,
                token=token,
            )

        git_instance_mock.push.assert_has_calls(
            [
                call(follow_tags=True, remote="upstream"),
                call(follow_tags=True, remote="upstream"),
            ],
        )

        create_api_mock.assert_awaited_once_with(
            "bar/foo",
            "v0.0.1",
            name="foo 0.0.1",
            body="A Changelog",
            prerelease=False,
        )

        command_mock.update_version.assert_has_calls(
            [
                call(release_version, force=False),
                call(next_version, force=False),
            ],
        )

        git_instance_mock.add.assert_called_with("MyProject.conf")
        git_instance_mock.commit.assert_called_with(
            "Automatic adjustments after release\n\n"
            "* Update to version 0.0.2.dev1\n",
            verify=False,
            gpg_signing_key="1234",
        )
        git_instance_mock.tag.assert_called_once_with(
            "v0.0.1", gpg_key_id="1234", message="Automatic release to 0.0.1"
        )

        self.assertEqual(released, CreateReleaseReturnValue.SUCCESS)

    @patch("pontos.release.create.Git", autospec=True)
    @patch("pontos.release.create.get_last_release_version", autospec=True)
    @patch("pontos.release.create.GitHubAsyncRESTApi", autospec=True)
    @patch(
        "pontos.release.create.CreateReleaseCommand._create_changelog",
        autospec=True,
    )
    @patch("pontos.release.create.Project._gather_commands", autospec=True)
    def test_release_github_api_pre_release(
        self,
        gather_commands_mock: MagicMock,
        create_changelog_mock: MagicMock,
        github_api_mock: AsyncMock,
        get_last_release_version_mock: MagicMock,
        git_mock,
    ):
        current_version = PEP440Version("0.0.0")
        release_version = PEP440Version("0.0.1a1")
        next_version = PEP440Version("0.0.1a1+dev1")
        command_mock = MagicMock(spec=GoVersionCommand)
        gather_commands_mock.return_value = [command_mock]
        get_last_release_version_mock.return_value = current_version
        command_mock.update_version.side_effect = [
            VersionUpdate(
                previous=current_version,
                new=release_version,
                changed_files=["MyProject.conf"],
            ),
            VersionUpdate(
                previous=release_version,
                new=next_version,
                changed_files=["MyProject.conf"],
            ),
        ]
        create_changelog_mock.return_value = "A Changelog"
        create_api_mock = AsyncMock()
        github_api_mock.return_value.releases.create = create_api_mock
        git_instance_mock: MagicMock = git_mock.return_value
        git_instance_mock.status.return_value = [
            StatusEntry("M  MyProject.conf")
        ]

        _, token, args = parse_args(
            [
                "release",
                "--space",
                "bar",
                "--project",
                "foo",
                "--release-version",
                "0.0.1a1",
                "--next-version",
                "0.0.1a1+dev1",
                "--git-remote-name",
                "upstream",
            ]
        )

        with temp_git_repository():
            released = create_release(
                terminal=mock_terminal(),
                error_terminal=mock_terminal(),
                args=args,
                token=token,
            )

        git_instance_mock.push.assert_has_calls(
            [
                call(follow_tags=True, remote="upstream"),
                call(follow_tags=True, remote="upstream"),
            ],
        )

        create_api_mock.assert_awaited_once_with(
            "bar/foo",
            "v0.0.1a1",
            name="foo 0.0.1a1",
            body="A Changelog",
            prerelease=True,
        )

        command_mock.update_version.assert_has_calls(
            [
                call(release_version, force=False),
                call(next_version, force=False),
            ],
        )

        git_instance_mock.add.assert_called_with("MyProject.conf")
        git_instance_mock.commit.assert_called_with(
            "Automatic adjustments after release\n\n"
            "* Update to version 0.0.1a1+dev1\n",
            verify=False,
            gpg_signing_key="1234",
        )
        git_instance_mock.tag.assert_called_once_with(
            "v0.0.1a1",
            gpg_key_id="1234",
            message="Automatic release to 0.0.1a1",
        )

        self.assertEqual(released, CreateReleaseReturnValue.SUCCESS)

    @patch("pontos.release.create.Git", autospec=True)
    @patch(
        "pontos.release.create.CreateReleaseCommand._create_release",
        autospec=True,
    )
    @patch(
        "pontos.release.create.CreateReleaseCommand._create_changelog",
        autospec=True,
    )
    def test_release_with_go_project(
        self,
        create_changelog_mock: MagicMock,
        create_release_mock: AsyncMock,
        git_mock,
    ):
        release_version = PEP440Version("0.0.2")

        create_changelog_mock.return_value = "A Changelog"

        git_instance_mock: MagicMock = git_mock.return_value
        git_instance_mock.status.return_value = [
            StatusEntry(f"M  {GoVersionCommand.version_file_path}")
        ]

        _, token, args = parse_args(["release", "--release-type", "patch"])

        with setup_go_project(current_version="0.0.1", tags=["0.0.1"]):
            released = create_release(
                terminal=mock_terminal(),
                error_terminal=mock_terminal(),
                args=args,
                token=token,
            )

        self.assertEqual(released, CreateReleaseReturnValue.SUCCESS)

        git_instance_mock.push.assert_called_with(follow_tags=True, remote=None)
        git_instance_mock.add.assert_called_with(
            GoVersionCommand.version_file_path
        )
        git_instance_mock.commit.assert_called_with(
            "Automatic adjustments after release\n\n"
            "* Update to version 0.0.3.dev1\n",
            verify=False,
            gpg_signing_key="1234",
        )
        git_instance_mock.tag.assert_called_once_with(
            "v0.0.2", gpg_key_id="1234", message="Automatic release to 0.0.2"
        )

        self.assertEqual(
            create_release_mock.await_args.args[1:],
            (release_version, "foo", "A Changelog"),
        )

        self.assertEqual(released, CreateReleaseReturnValue.SUCCESS)

    @patch("pontos.release.create.Git", autospec=True)
    @patch("pontos.release.create.get_last_release_version", autospec=True)
    @patch("pontos.changelog.conventional_commits.Git", autospec=True)
    @patch(
        "pontos.release.create.CreateReleaseCommand._create_release",
        autospec=True,
    )
    @patch("pontos.release.create.Project._gather_commands", autospec=True)
    def test_release_with_changelog(
        self,
        gather_commands_mock: MagicMock,
        create_release_mock: AsyncMock,
        cc_git_mock: MagicMock,
        get_last_release_version_mock: MagicMock,
        git_mock: MagicMock,
    ):
        current_version = PEP440Version("0.0.1")
        release_version = PEP440Version("0.0.2")
        next_version = PEP440Version("1.0.0.dev1")
        command_mock = MagicMock(spec=GoVersionCommand)
        gather_commands_mock.return_value = [command_mock]
        get_last_release_version_mock.return_value = current_version
        command_mock.update_version.side_effect = [
            VersionUpdate(
                previous=current_version,
                new=release_version,
                changed_files=["project.conf"],
            ),
            VersionUpdate(
                previous=release_version,
                new=next_version,
                changed_files=["project.conf", "version.lang"],
            ),
        ]
        git_instance_mock: MagicMock = git_mock.return_value
        git_instance_mock.status.return_value = [StatusEntry("M  project.conf")]
        cc_git_mock.return_value.log.return_value = [
            "1234567 Add: foo bar",
            "8abcdef Add: bar baz",
            "8abcd3f Add bar baz",
            "8abcd3d Adding bar baz",
            "1337abc Change: bar to baz",
            "42a42a4 Remove: foo bar again",
            "fedcba8 Test: bar baz testing",
            "dead901 Refactor: bar baz ref",
            "fedcba8 Fix: bar baz fixing",
            "d0c4d0c Doc: bar baz documenting",
        ]
        today = datetime.today().strftime("%Y-%m-%d")
        expected_changelog = f"""## [0.0.2] - {today}

## Added
* foo bar [1234567](https://github.com/greenbone/foo/commit/1234567)
* bar baz [8abcdef](https://github.com/greenbone/foo/commit/8abcdef)

## Removed
* foo bar again [42a42a4](https://github.com/greenbone/foo/commit/42a42a4)

## Changed
* bar to baz [1337abc](https://github.com/greenbone/foo/commit/1337abc)

## Bug Fixes
* bar baz fixing [fedcba8](https://github.com/greenbone/foo/commit/fedcba8)

[0.0.2]: https://github.com/greenbone/foo/compare/v0.0.1...v0.0.2"""

        _, token, args = parse_args(
            [
                "release",
                "--project",
                "foo",
                "--release-version",
                "0.0.2",
                "--next-version",
                "1.0.0.dev1",
            ]
        )

        with temp_git_repository():
            released = create_release(
                terminal=mock_terminal(),
                error_terminal=mock_terminal(),
                args=args,
                token=token,
            )

        git_instance_mock.list_tags.assert_called_once_with()

        cc_git_mock.return_value.log.assert_called_once_with(
            "v0.0.1..HEAD", oneline=True
        )

        git_instance_mock.push.assert_has_calls(
            [
                call(follow_tags=True, remote=None),
                call(follow_tags=True, remote=None),
            ]
        )

        command_mock.update_version.assert_has_calls(
            [
                call(release_version, force=False),
                call(next_version, force=False),
            ]
        )

        self.assertEqual(
            create_release_mock.await_args.args[1:],
            (release_version, "foo", expected_changelog),
        )

        git_instance_mock.tag.assert_called_once_with(
            "v0.0.2", gpg_key_id="1234", message="Automatic release to 0.0.2"
        )
        git_instance_mock.add.assert_has_calls(
            [call("project.conf"), call("project.conf"), call("version.lang")]
        )
        git_instance_mock.commit.assert_has_calls(
            [
                call(
                    "Automatic release to 0.0.2",
                    verify=False,
                    gpg_signing_key="1234",
                ),
                call(
                    "Automatic adjustments after release\n\n"
                    "* Update to version 1.0.0.dev1\n",
                    verify=False,
                    gpg_signing_key="1234",
                ),
            ]
        )

        self.assertEqual(released, CreateReleaseReturnValue.SUCCESS)

    @patch("pontos.release.create.Git", autospec=True)
    @patch("pontos.release.create.get_last_release_version", autospec=True)
    @patch(
        "pontos.release.create.CreateReleaseCommand._create_release",
        autospec=True,
    )
    @patch(
        "pontos.release.create.CreateReleaseCommand._create_changelog",
        autospec=True,
    )
    @patch("pontos.release.create.Project._gather_commands", autospec=True)
    def test_release_local(
        self,
        gather_commands_mock: MagicMock,
        create_changelog_mock: MagicMock,
        create_release_mock: AsyncMock,
        get_last_release_version_mock: MagicMock,
        git_mock,
    ):
        current_version = PEP440Version("0.0.0")
        release_version = PEP440Version("0.0.1")
        next_version = PEP440Version("0.0.2.dev1")
        command_mock = MagicMock(spec=GoVersionCommand)
        gather_commands_mock.return_value = [command_mock]
        get_last_release_version_mock.return_value = current_version
        command_mock.update_version.side_effect = [
            VersionUpdate(
                previous=current_version,
                new=release_version,
                changed_files=["MyProject.conf"],
            ),
            VersionUpdate(
                previous=release_version,
                new=next_version,
                changed_files=["MyProject.conf"],
            ),
        ]
        create_changelog_mock.return_value = "A Changelog"
        git_instance_mock: MagicMock = git_mock.return_value
        git_instance_mock.status.return_value = [
            StatusEntry("M  MyProject.conf")
        ]

        _, token, args = parse_args(
            [
                "release",
                "--project",
                "foo",
                "--release-version",
                "0.0.1",
                "--next-version",
                "0.0.2.dev1",
                "--local",
            ]
        )

        with temp_git_repository():
            released = create_release(
                terminal=mock_terminal(),
                error_terminal=mock_terminal(),
                args=args,
                token=token,
            )

        git_instance_mock.push.assert_not_called()

        create_release_mock.assert_not_awaited()

        command_mock.update_version.assert_has_calls(
            [
                call(release_version, force=False),
                call(next_version, force=False),
            ]
        )

        git_instance_mock.add.assert_called_with("MyProject.conf")
        git_instance_mock.commit.assert_called_with(
            "Automatic adjustments after release\n\n"
            "* Update to version 0.0.2.dev1\n",
            verify=False,
            gpg_signing_key="1234",
        )
        git_instance_mock.tag.assert_called_once_with(
            "v0.0.1", gpg_key_id="1234", message="Automatic release to 0.0.1"
        )

        self.assertEqual(released, CreateReleaseReturnValue.SUCCESS)


@dataclass
class Release:
    release_type: str
    expected_release_version: str
    tags: list[str] = field(default_factory=list)
    expected_last_release_version: Optional[str] = None
    current_version: Optional[str] = None
    release_series: Optional[str] = None


@patch.dict(
    "os.environ",
    {"GITHUB_TOKEN": "foo", "GITHUB_USER": "user", "GPG_SIGNING_KEY": ""},
)
class ReleaseGoProjectTestCase(unittest.TestCase):
    @patch("pontos.release.create.Git.push", autospec=True)
    @patch(
        "pontos.release.create.GitHubAsyncRESTApi",
        autospec=True,
    )
    def test_release(
        self, github_api_mock: AsyncMock, _git_push_mock: MagicMock
    ):
        create_api_mock = AsyncMock()
        github_api_mock.return_value.releases.create = create_api_mock

        releases = [
            Release(
                release_type="major",
                current_version="1.0.0",
                expected_release_version="2.0.0",
                tags="1.0.0",
            ),
            Release(
                release_type="minor",
                current_version="1.0.0",
                expected_release_version="1.1.0",
                tags="1.0.0",
            ),
            Release(
                release_type="patch",
                current_version="1.0.0",
                expected_release_version="1.0.1",
                tags="1.0.0",
            ),
            Release(
                release_type="patch",
                current_version="1.0.5",
                expected_release_version="1.0.6",
                tags="1.0.5",
            ),
            Release(
                release_type="alpha",
                current_version="1.0.0",
                expected_release_version="1.0.1a1",
                tags="1.0.0",
            ),
            Release(
                release_type="alpha",
                current_version="1.0.0a3",
                expected_release_version="1.0.0a4",
                tags="1.0.0a3",
            ),
            Release(
                release_type="beta",
                current_version="1.0.0",
                expected_release_version="1.0.1b1",
                tags="1.0.0",
            ),
            Release(
                release_type="beta",
                current_version="1.0.0b2",
                expected_release_version="1.0.0b3",
                tags="1.0.0b2",
            ),
            Release(
                release_type="release-candidate",
                current_version="1.0.0",
                expected_release_version="1.0.1rc1",
                tags="1.0.0",
            ),
            Release(
                release_type="release-candidate",
                current_version="1.0.0rc1",
                expected_release_version="1.0.0rc2",
                tags="1.0.0rc1",
            ),
        ]

        for r in releases:
            with setup_go_project(
                current_version=r.current_version, tags=r.tags
            ):
                _, token, args = parse_args(
                    ["release", "--release-type", r.release_type]
                )
                released = create_release(
                    terminal=mock_terminal(),
                    error_terminal=mock_terminal(),
                    args=args,
                    token=token,
                )

            self.assertEqual(
                released,
                CreateReleaseReturnValue.SUCCESS,
                f"v{r.expected_release_version}",
            )
            self.assertEqual(
                create_api_mock.call_args.args[1],
                f"v{r.expected_release_version}",
            )

            create_api_mock.reset_mock()

    @patch(
        "pontos.release.create.CreateReleaseCommand._create_changelog",
        autospec=True,
    )
    @patch("pontos.release.create.Git.push", autospec=True)
    @patch(
        "pontos.release.create.GitHubAsyncRESTApi",
        autospec=True,
    )
    def test_release_series(
        self,
        github_api_mock: AsyncMock,
        _git_push_mock: MagicMock,
        create_changelog_mock: MagicMock,
    ):
        create_api_mock = AsyncMock()
        github_api_mock.return_value.releases.create = create_api_mock

        create_changelog_mock.return_value = "A Changelog"

        releases = [
            Release(
                release_type="major",
                current_version="1.0.0",
                expected_release_version="2.0.0",
                expected_last_release_version="1.0.0",
                tags=["1.0.0", "3.0.0"],
                release_series="1",
            ),
            Release(
                release_type="major",
                current_version="1.0.0rc1",
                expected_release_version="1.0.0",
                expected_last_release_version="0.9.0",
                tags=["0.9.0", "0.8.1", "0.5.0"],
            ),
            Release(
                release_type="minor",
                current_version="1.0.0",
                expected_release_version="2.1.0",
                expected_last_release_version="2.0.0",
                tags=["1.0.0", "2.0.0"],
            ),
            Release(
                release_type="minor",
                current_version="1.0.0",
                expected_release_version="1.1.0",
                expected_last_release_version="1.0.0",
                tags=["1.0.0", "2.0.0"],
                release_series="1",
            ),
            Release(
                release_type="minor",
                current_version="1.1.0",
                expected_release_version="1.2.0",
                expected_last_release_version="1.1.0",
                tags=["1.0.0", "1.1.0", "2.0.0"],
                release_series="1.1",
            ),
            Release(
                release_type="patch",
                current_version="1.0.0",
                expected_release_version="1.0.1",
                expected_last_release_version="1.0.0",
                tags=["1.0.0", "1.1.0", "2.0.0"],
                release_series="1.0",
            ),
            Release(
                release_type="patch",
                current_version="1.0.5",
                expected_release_version="1.0.6",
                expected_last_release_version="1.0.5",
                tags=["1.0.5", "1.1.0", "2.0.0"],
                release_series="1.0",
            ),
            Release(
                release_type="patch",
                current_version="1.0.5",
                expected_release_version="1.1.1",
                expected_last_release_version="1.1.0",
                tags=["1.0.5", "1.1.0", "2.0.0"],
                release_series="1.1",
            ),
            Release(
                release_type="patch",
                current_version="1.1.0rc1",
                expected_release_version="1.1.0",
                expected_last_release_version="1.1.0b1",
                tags=["1.0.0", "1.1.0b1", "1.2.0", "2.0.0"],
                release_series="1.1",
            ),
            Release(
                release_type="alpha",
                current_version="1.0.0",
                expected_release_version="1.0.1a1",
                expected_last_release_version="1.0.0",
                tags=["1.0.0", "2.0.0", "1.1.0"],
                release_series="1.0",
            ),
            Release(
                release_type="alpha",
                current_version="1.0.0a3",
                expected_release_version="1.0.0a4",
                expected_last_release_version="1.0.0a3",
                tags=["1.0.0a3", "2.0.0", "1.1.0"],
                release_series="1.0",
            ),
            Release(
                release_type="beta",
                current_version="1.0.0",
                expected_release_version="1.0.1b1",
                expected_last_release_version="1.0.0",
                tags=["1.0.0", "2.0.0", "1.1.0"],
                release_series="1.0",
            ),
            Release(
                release_type="beta",
                current_version="1.0.0b2",
                expected_release_version="1.0.0b3",
                expected_last_release_version="1.0.0b2",
                tags=["1.0.0b2", "2.0.0", "1.1.0"],
                release_series="1.0",
            ),
            Release(
                release_type="release-candidate",
                current_version="1.0.0",
                expected_release_version="1.0.1rc1",
                expected_last_release_version="1.0.0",
                tags=["1.0.0", "2.0.0", "1.1.0"],
                release_series="1.0",
            ),
            Release(
                release_type="release-candidate",
                current_version="1.0.0rc1",
                expected_release_version="1.0.0rc2",
                expected_last_release_version="1.0.0rc1",
                tags=["1.0.0rc1", "2.0.0", "1.1.0"],
                release_series="1.0",
            ),
        ]

        for r in releases:
            with setup_go_project(
                current_version=r.current_version, tags=r.tags
            ):
                input_args = ["release", "--release-type", r.release_type]
                if r.release_series:
                    input_args.extend(["--release-series", r.release_series])

                _, token, args = parse_args(input_args)
                released = create_release(
                    terminal=mock_terminal(),
                    error_terminal=mock_terminal(),
                    args=args,
                    token=token,
                )

            self.assertEqual(released, CreateReleaseReturnValue.SUCCESS)
            self.assertEqual(
                create_api_mock.call_args.args[1],
                f"v{r.expected_release_version}",
                f"Unexpected current release version {r}",
            )

            self.assertEqual(
                create_changelog_mock.call_args.args[1],
                PEP440Version.from_string(r.expected_release_version),
                f"Unexpected current release version {r}",
            )

            if r.expected_last_release_version is None:
                self.assertIsNone(create_changelog_mock.call_args.args[2])
            else:
                self.assertEqual(
                    create_changelog_mock.call_args.args[2],
                    PEP440Version.from_string(r.expected_last_release_version),
                    f"Unexpected last release version for {r}",
                )

            create_api_mock.reset_mock()
            create_changelog_mock.reset_mock()



# -*- coding: utf-8 -*-
# Copyright (C) 2021-2022 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from typing import Iterator, Optional

from pontos.git import Git, TagSort
from pontos.git.git import DEFAULT_TAG_SORT_SUFFIX

from .version import ParseVersionFuncType, Version


def get_last_release_versions(
    parse_version: ParseVersionFuncType,
    *,
    git_tag_prefix: Optional[str] = "",
    ignore_pre_releases: Optional[bool] = False,
    tag_name: Optional[str] = None,
) -> Iterator[Version]:
    """Get the last released Versions from git.

    Args:
        git_tag_prefix: Git tag prefix to consider
        ignore_pre_release: Ignore pre releases and only consider non pre
            releases. Default is False.
        tag_name: A pattern for filtering the tags. For example: "1.2.*"

    Returns:
        List of released versions
    """

    tag_list = Git().list_tags(
        sort=TagSort.VERSION,
        sort_suffix=DEFAULT_TAG_SORT_SUFFIX,
        tag_name=tag_name,
    )
    tag_list.reverse()

    for tag in tag_list:
        last_release_version = tag.strip(git_tag_prefix)

        version = parse_version(last_release_version)
        if version.is_pre_release and ignore_pre_releases:
            continue

        yield version


def get_last_release_version(
    parse_version: ParseVersionFuncType,
    *,
    git_tag_prefix: Optional[str] = "",
    ignore_pre_releases: Optional[bool] = False,
    tag_name: Optional[str] = None,
) -> Optional[Version]:
    """Get the last released Version from git.

    Args:
        git_tag_prefix: Git tag prefix to consider
        ignore_pre_release: Ignore pre releases and only consider non pre
            releases. Default is False.
        tag_name: A pattern for filtering the tags. For example: "1.2.*"

    Returns:
        Last released git tag as Version if tags were found
        or None
    """

    it = get_last_release_versions(
        parse_version=parse_version,
        git_tag_prefix=git_tag_prefix,
        ignore_pre_releases=ignore_pre_releases,
        tag_name=tag_name,
    )

    try:
        return next(it)
    except StopIteration:
        return None



# Copyright (C) 2023 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# pylint: disable=line-too-long
# ruff: noqa: E501

import unittest
from pathlib import Path

from lxml import etree

from pontos.testing import temp_directory, temp_file
from pontos.version import VersionError
from pontos.version.commands import JavaVersionCommand
from pontos.version.commands._java import (
    VERSION_PATTERN,
    find_file,
    replace_string_in_file,
)
from pontos.version.schemes import PEP440VersioningScheme


class TestVersionPattern(unittest.TestCase):
    def test_version_pattern_swagger(self):
        # Define the test parameters
        content = """
@Configuration
public class SwaggerConfig {{

    @Bean
    public OpenAPI metaData() {{
        return new OpenAPI().info(new Info()
                .title("Api Documentation")
                .description("Api Documentation for billing service")
                .version("{}"));
    }}
}}
        """
        versions = [
            "2018.0.1",
            "2023.1.2a1",
            "2023.3.3-alpha1.dev1",
            "2023.4.5-rc2",
            "2023.1.1beta1",
        ]
        replacement = "2023.10.10"

        for version in versions:
            with temp_file(content=content.format(version)) as tmp:
                replace_string_in_file(tmp, VERSION_PATTERN, replacement)

                updated_content = tmp.read_text(encoding="utf-8")

                # Verify the replacement was performed correctly
                self.assertNotRegex(
                    updated_content, version
                )  # Pattern should not be present
                self.assertIn(
                    replacement, updated_content
                )  # Replacement should be present

    def test_version_pattern_properties(self):
        # Define the test parameters
        content = """
# application
spring.application.name=boo
server.port=1elf
app.stripe.enabled=false
app.not.for.resale.keys=
app.gmsp.booking.startOfHistory=2021
# sentry
sentry.release={}
sentry.tags.service_name=boo
# actuator
management.health.db.enabled=false
# spring
spring.main.allow-bean-definition-overriding=true
        """
        versions = [
            "2018.0.1",
            "2023.1.2a1",
            "2023.3.3-alpha1.dev1",
            "2023.4.5-rc2",
            "2023.1.1beta1",
        ]
        replacement = "2023.10.10"

        for version in versions:
            with temp_file(content=content.format(version)) as tmp:
                replace_string_in_file(tmp, VERSION_PATTERN, replacement)

                updated_content = tmp.read_text(encoding="utf-8")

                # Verify the replacement was performed correctly
                self.assertNotRegex(
                    updated_content, version
                )  # Pattern should not be present
                self.assertIn(
                    replacement, updated_content
                )  # Replacement should be present


class TestFindFile(unittest.TestCase):
    def test_file_found(self):
        with temp_directory() as temp_dir:
            deep_path = Path(temp_dir / "foo/bat/config/swagger/")
            deep_path.mkdir(parents=True)
            Path(deep_path / "SwaggerConfig.java").touch()
            self.assertTrue(
                Path(
                    temp_dir / "foo/bat/config/swagger/SwaggerConfig.java"
                ).exists()
            )
            filename = Path("SwaggerConfig.java")
            search_path = temp_dir  # Assuming 'config/swagger' is two levels up
            search_glob = "**/config/swagger/*"

            result = find_file(filename, search_path, search_glob)

            self.assertIsNotNone(result)
            self.assertEqual(result.name, filename.name)

    def test_file_not_found(self):
        with temp_directory() as temp_dir:
            Path(
                temp_dir / "foo/bat/config/swagger/SwaggerConfig.java",
                parents=True,
            )
            filename = Path("NonExistentFile.java")
            search_path = temp_dir
            search_glob = "**/config/swagger/*"
            with self.assertLogs("root", level="WARNING") as cm:
                result = find_file(filename, search_path, search_glob)
                self.assertEqual(
                    cm.output,
                    [
                        (
                            f"WARNING:root:File {filename.name} not "
                            f"found in {search_path.resolve()}."
                        )
                    ],
                )

            self.assertIsNone(result)


class TestReplaceString(unittest.TestCase):
    def test_replace_string_in_file(self):
        # Define the test parameters
        content = """
        Foo, bar
        baz
            .version("1.2.3")
        glubb
        """
        pattern = r'\.version\("([0-9]+\.[0-9]+\.[0-9]+)"\)'
        replacement = "1.2.4"

        with temp_file(content=content) as tmp:
            replace_string_in_file(tmp, pattern, replacement)

            updated_content = tmp.read_text(encoding="utf-8")

            # Verify the replacement was performed correctly
            self.assertNotRegex(
                updated_content, "1.2.3"
            )  # Pattern should not be present
            self.assertIn(
                replacement, updated_content
            )  # Replacement should be present

    def test_replace_string_in_file_no_match(self):
        # Define the test parameters
        content = """
        Foo, bar
        baz
            .versio("1.2.3")
        glubb
        """
        pattern = r'\.version\("([0-9]+\.[0-9]+\.[0-9]+)"\)'
        replacement = "1.2.4"

        with temp_file(content=content) as tmp:
            # Call the function under test
            replace_string_in_file(tmp, pattern, replacement)

            # Read the content of the unmodified file
            updated_content = tmp.read_text(encoding="utf-8")

            # Verify the content remains unchanged
            self.assertNotRegex(updated_content, replacement)
            self.assertEqual(updated_content, content)


class GetCurrentJavaVersionCommandTestCase(unittest.TestCase):
    def test_get_current_version(self):
        content = """<?xml version="1.0" encoding="UTF-8"?>
        <project xmlns="http://maven.apache.org/POM/4.0.0" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:schemaLocation="http://maven.apache.org/POM/4.0.0 https://maven.apache.org/xsd/maven-4.0.0.xsd">
        <modelVersion>4.0.0</modelVersion><groupId>net.greenbone.umbrella</groupId>
        <artifactId>msspadminservice</artifactId><version>2023.5.3</version></project>"""
        with temp_file(content, name="pom.xml", change_into=True):
            cmd = JavaVersionCommand(PEP440VersioningScheme)
            version = cmd.get_current_version()

            self.assertEqual(
                version, PEP440VersioningScheme.parse_version("2023.5.3")
            )

    def test_no_project_file(self):
        with temp_directory(change_into=True), self.assertRaisesRegex(
            VersionError, ".* file not found."
        ):
            cmd = JavaVersionCommand(PEP440VersioningScheme)
            cmd.get_current_version()

    def test_no_package_version(self):
        content = """<?xml version="1.0" encoding="UTF-8"?>
        <project xmlns="http://maven.apache.org/POM/4.0.0" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
                xsi:schemaLocation="http://maven.apache.org/POM/4.0.0 https://maven.apache.org/xsd/maven-4.0.0.xsd">
            <modelVersion>4.0.0</modelVersion><groupId>net.greenbone.umbrella</groupId>
            <artifactId>msspadminservice</artifactId></project>"""
        with temp_file(
            content, name="pom.xml", change_into=True
        ), self.assertRaisesRegex(VersionError, "Version tag missing in"):
            cmd = JavaVersionCommand(PEP440VersioningScheme)
            cmd.get_current_version()

    def test_no_valid_xml_in_pom(self):
        content = "<"
        with temp_file(
            content, name="pom.xml", change_into=True
        ), self.assertRaisesRegex(
            VersionError, "StartTag: invalid element name,"
        ):
            cmd = JavaVersionCommand(PEP440VersioningScheme)
            cmd.get_current_version()

        content = "<foo><bar/>"
        with temp_file(
            content, name="pom.xml", change_into=True
        ), self.assertRaisesRegex(
            VersionError, "Premature end of data in tag foo"
        ):
            cmd = JavaVersionCommand(PEP440VersioningScheme)
            cmd.get_current_version()


class UpdateJavaVersionCommandTestCase(unittest.TestCase):
    swagger_file = "SwaggerConfig.java"

    def test_update_version_file(self):
        content = """<?xml version="1.0" encoding="UTF-8"?>
        <project xmlns="http://maven.apache.org/POM/4.0.0" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:schemaLocation="http://maven.apache.org/POM/4.0.0 https://maven.apache.org/xsd/maven-4.0.0.xsd">
        <modelVersion>4.0.0</modelVersion><groupId>net.greenbone.umbrella</groupId>
        <artifactId>msspadminservice</artifactId><version>2023.5.3</version></project>"""

        with temp_file(content, name="pom.xml", change_into=True) as temp:
            cmd = JavaVersionCommand(PEP440VersioningScheme)
            cmd.get_current_version()
            updated = cmd.update_version(
                PEP440VersioningScheme.parse_version("2023.6.0")
            )

            self.assertEqual(
                updated.previous,
                PEP440VersioningScheme.parse_version("2023.5.3"),
            )
            self.assertEqual(
                updated.new, PEP440VersioningScheme.parse_version("2023.6.0")
            )
            self.assertEqual(updated.changed_files, [temp.resolve()])

            with temp.open(mode="r", encoding="utf-8") as fp:
                fake_pom = etree.parse(fp).getroot()

            self.assertEqual(fake_pom.find("{*}version").text, "2023.6.0")

    def test_update_version_develop(self):
        content = """<?xml version="1.0" encoding="UTF-8"?>
        <project xmlns="http://maven.apache.org/POM/4.0.0" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:schemaLocation="http://maven.apache.org/POM/4.0.0 https://maven.apache.org/xsd/maven-4.0.0.xsd">
        <modelVersion>4.0.0</modelVersion><groupId>net.greenbone.umbrella</groupId>
        <artifactId>msspadminservice</artifactId><version>2023.5.3</version></project>"""

        with temp_file(content, name="pom.xml", change_into=True) as temp:
            swagger_search_path = Path("src").resolve()
            cmd = JavaVersionCommand(PEP440VersioningScheme)
            new_version = PEP440VersioningScheme.parse_version("2023.6.0.dev1")
            with self.assertLogs("root", level="WARNING") as cm:
                updated = cmd.update_version(new_version)
                self.assertEqual(
                    cm.output,
                    [
                        (
                            f"WARNING:root:File {self.swagger_file} not "
                            f"found in {swagger_search_path.resolve()}."
                        )
                    ],
                )

            self.assertEqual(
                updated.previous,
                PEP440VersioningScheme.parse_version("2023.5.3"),
            )
            self.assertEqual(
                updated.new,
                PEP440VersioningScheme.parse_version("2023.6.0.dev1"),
            )
            self.assertEqual(updated.changed_files, [temp.resolve()])

            with temp.open(mode="r", encoding="utf-8") as fp:
                fake_pom = etree.parse(fp).getroot()

            self.assertEqual(fake_pom.find("{*}version").text, "2023.6.0.dev1")

    def test_no_update(self):
        content = """<?xml version="1.0" encoding="UTF-8"?>
        <project xmlns="http://maven.apache.org/POM/4.0.0" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:schemaLocation="http://maven.apache.org/POM/4.0.0 https://maven.apache.org/xsd/maven-4.0.0.xsd">
        <modelVersion>4.0.0</modelVersion><groupId>net.greenbone.umbrella</groupId>
        <artifactId>msspadminservice</artifactId><version>2023.5.3</version></project>"""

        with temp_file(content, name="pom.xml", change_into=True) as temp:
            cmd = JavaVersionCommand(PEP440VersioningScheme)
            updated = cmd.update_version(
                PEP440VersioningScheme.parse_version("2023.5.3")
            )

            self.assertEqual(
                updated.previous,
                PEP440VersioningScheme.parse_version("2023.5.3"),
            )
            self.assertEqual(
                updated.new, PEP440VersioningScheme.parse_version("2023.5.3")
            )
            self.assertEqual(updated.changed_files, [])

            with temp.open(mode="r", encoding="utf-8") as fp:
                fake_pom = etree.parse(fp).getroot()

            self.assertEqual(fake_pom.find("{*}version").text, "2023.5.3")

    def test_forced_update(self):
        content = """<?xml version="1.0" encoding="UTF-8"?>
        <project xmlns="http://maven.apache.org/POM/4.0.0" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:schemaLocation="http://maven.apache.org/POM/4.0.0 https://maven.apache.org/xsd/maven-4.0.0.xsd">
        <modelVersion>4.0.0</modelVersion><groupId>net.greenbone.umbrella</groupId>
        <artifactId>msspadminservice</artifactId><version>2023.5.3</version></project>"""

        with temp_file(content, name="pom.xml", change_into=True) as temp:
            cmd = JavaVersionCommand(PEP440VersioningScheme)
            updated = cmd.update_version(
                PEP440VersioningScheme.parse_version("2023.5.3"), force=True
            )

            self.assertEqual(
                updated.previous,
                PEP440VersioningScheme.parse_version("2023.5.3"),
            )
            self.assertEqual(
                updated.new, PEP440VersioningScheme.parse_version("2023.5.3")
            )
            self.assertEqual(updated.changed_files, [temp.resolve()])

            with temp.open(mode="r", encoding="utf-8") as fp:
                fake_pom = etree.parse(fp).getroot()

            self.assertEqual(fake_pom.find("{*}version").text, "2023.5.3")



# Copyright (C) 2022 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
This script prints out all artifacts of a given repository
"""

from argparse import ArgumentParser, Namespace

from rich.console import Console

from pontos.github.api import GitHubAsyncRESTApi


def add_script_arguments(parser: ArgumentParser) -> None:
    parser.add_argument("repository")


async def github_script(api: GitHubAsyncRESTApi, args: Namespace) -> int:
    count = 0
    expired = 0
    size = 0.0
    async for artifact in api.artifacts.get_all(args.repository):
        count += 1
        if artifact.expired:
            expired += 1
            continue

        size += artifact.size_in_bytes / (1024 * 1024)

    console = Console()
    console.print(f"{count} artifacts.")
    console.print(f"{expired} expired.")
    console.print(f"Size {size:.2f} MiB")

    return 0



# Copyright (C) 2020-2022 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys
from argparse import ArgumentParser, Namespace
from pathlib import Path
from typing import NoReturn, Optional, Sequence

from pontos.changelog.conventional_commits import ChangelogBuilder
from pontos.errors import PontosError
from pontos.terminal.null import NullTerminal
from pontos.terminal.rich import RichTerminal
from pontos.version.helper import get_last_release_version
from pontos.version.schemes import (
    VERSIONING_SCHEMES,
    VersioningScheme,
    versioning_scheme_argument_type,
)


def parse_args(args: Optional[Sequence[str]] = None) -> Namespace:
    parser = ArgumentParser(
        description="Conventional commits utility.",
        prog="pontos-changelog",
    )

    parser.add_argument(
        "--config",
        "-C",
        type=Path,
        help="Conventional commits config file (toml), including conventions.",
    )

    parser.add_argument(
        "--project",
        required=True,
        help="The github project",
    )

    parser.add_argument(
        "--space",
        default="greenbone",
        help="User/Team name in github",
    )

    parser.add_argument(
        "--versioning-scheme",
        help="Versioning scheme to use for parsing and handling version "
        f"information. Choices are {', '.join(VERSIONING_SCHEMES.keys())}. "
        "Default: %(default)s",
        default="pep440",
        type=versioning_scheme_argument_type,
    )

    parser.add_argument(
        "--current-version",
        help="Current version before these changes",
    )

    parser.add_argument(
        "--next-version",
        help="The planned release version",
    )

    parser.add_argument(
        "--git-tag-prefix",
        default="v",
        help="Prefix for git tag versions. Default: %(default)s",
    )

    parser.add_argument(
        "--output",
        "-o",
        type=Path,
        help="Write changelog to this file.",
    )

    parser.add_argument(
        "--quiet",
        "-q",
        action="store_true",
        help="Don't print messages to the terminal",
    )

    parsed_args = parser.parse_args(args=args)

    scheme: VersioningScheme = parsed_args.versioning_scheme
    current_version = getattr(parsed_args, "current_version", None)
    if current_version:
        parsed_args.current_version = scheme.parse_version(current_version)

    next_version = getattr(parsed_args, "next_version", None)
    if next_version:
        parsed_args.next_version = scheme.parse_version(next_version)

    return parsed_args


def main(args: Optional[Sequence[str]] = None) -> NoReturn:
    parsed_args = parse_args(args)

    term = NullTerminal() if parsed_args.quiet else RichTerminal()

    if parsed_args.current_version:
        last_version = parsed_args.current_version
    else:
        last_version = get_last_release_version(
            parsed_args.versioning_scheme.parse_version,
            git_tag_prefix=parsed_args.git_tag_prefix,
        )

    try:
        changelog_builder = ChangelogBuilder(
            config=parsed_args.config,
            project=parsed_args.project,
            space=parsed_args.space,
        )
        if parsed_args.output:
            changelog_builder.create_changelog_file(
                parsed_args.output,
                last_version=last_version,
                next_version=parsed_args.next_version,
            )
        else:
            changelog = changelog_builder.create_changelog(
                last_version=last_version,
                next_version=parsed_args.next_version,
            )
            term.out(changelog)
    except PontosError as e:
        term.error(str(e))
        sys.exit(1)

    sys.exit(0)


if __name__ == "__main__":
    main()



# Copyright (C) 2022 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# pylint: disable=line-too-long, arguments-differ, redefined-builtin
# ruff: noqa: E501

from datetime import datetime
from typing import Any, Dict, List, Optional
from unittest.mock import MagicMock, patch

from httpx import AsyncClient, Response

from pontos.errors import PontosError
from pontos.nvd.api import now, sleep
from pontos.nvd.cpe.api import CPEApi
from tests import AsyncMock, IsolatedAsyncioTestCase, aiter, anext
from tests.nvd import get_cpe_data


def create_cpe_response(
    cpe_name_id: str, update: Optional[Dict[str, Any]] = None
) -> MagicMock:
    data = {
        "products": [{"cpe": get_cpe_data({"cpe_name_id": cpe_name_id})}],
        "results_per_page": 1,
    }
    if update:
        data.update(update)

    response = MagicMock(spec=Response)
    response.json.return_value = data
    return response


def create_cpes_responses(count: int = 2) -> List[MagicMock]:
    return [
        create_cpe_response(f"CPE-{i}", {"total_results": count})
        for i in range(1, count + 1)
    ]


class CPEApiTestCase(IsolatedAsyncioTestCase):
    @patch("pontos.nvd.api.AsyncClient", spec=AsyncClient)
    def setUp(self, async_client: MagicMock) -> None:
        self.http_client = AsyncMock()
        async_client.return_value = self.http_client
        self.api = CPEApi()

    async def test_no_cpe_name_id(self):
        with self.assertRaises(PontosError):
            await self.api.cpe(None)

    async def test_no_cpe(self):
        data = {
            "products": [],
            "results_per_page": 1,
        }
        response = MagicMock(spec=Response)
        response.json.return_value = data
        self.http_client.get.return_value = response

        with self.assertRaises(PontosError):
            await self.api.cpe("CPE-1")

    async def test_cpe(self):
        self.http_client.get.return_value = create_cpe_response("CPE-1")

        cpe = await self.api.cpe("CPE-1")

        self.http_client.get.assert_awaited_once_with(
            "https://services.nvd.nist.gov/rest/json/cpes/2.0",
            headers={},
            params={"cpeNameId": "CPE-1"},
        )

        self.assertEqual(
            cpe.cpe_name,
            "cpe:2.3:o:microsoft:windows_10_22h2:-:*:*:*:*:*:arm64:*",
        )
        self.assertEqual(cpe.cpe_name_id, "CPE-1")
        self.assertFalse(cpe.deprecated)
        self.assertEqual(
            cpe.last_modified, datetime(2022, 12, 9, 18, 15, 16, 973000)
        )
        self.assertEqual(cpe.created, datetime(2022, 12, 9, 16, 20, 6, 943000))

        self.assertEqual(cpe.refs, [])
        self.assertEqual(cpe.titles, [])
        self.assertEqual(cpe.deprecated_by, [])

    @patch("pontos.nvd.api.sleep", spec=sleep)
    async def test_rate_limit(self, sleep_mock: MagicMock):
        self.http_client.get.side_effect = create_cpes_responses(6)

        it = aiter(self.api.cpes())
        await anext(it)
        await anext(it)
        await anext(it)
        await anext(it)
        await anext(it)

        sleep_mock.assert_not_called()

        await anext(it)

        sleep_mock.assert_called_once_with()

    @patch("pontos.nvd.cpe.api.now", spec=now)
    async def test_cves_last_modified_start_date(self, now_mock: MagicMock):
        now_mock.return_value = datetime(2022, 12, 31)
        self.http_client.get.side_effect = create_cpes_responses()

        it = aiter(
            self.api.cpes(last_modified_start_date=datetime(2022, 12, 1))
        )
        cve = await anext(it)

        self.assertEqual(cve.cpe_name_id, "CPE-1")
        self.http_client.get.assert_awaited_once_with(
            "https://services.nvd.nist.gov/rest/json/cpes/2.0",
            headers={},
            params={
                "startIndex": 0,
                "lastModStartDate": "2022-12-01T00:00:00",
                "lastModEndDate": "2022-12-31T00:00:00",
            },
        )

        self.http_client.get.reset_mock()

        cve = await anext(it)

        self.assertEqual(cve.cpe_name_id, "CPE-2")
        self.http_client.get.assert_awaited_once_with(
            "https://services.nvd.nist.gov/rest/json/cpes/2.0",
            headers={},
            params={
                "startIndex": 1,
                "lastModStartDate": "2022-12-01T00:00:00",
                "lastModEndDate": "2022-12-31T00:00:00",
                "resultsPerPage": 1,
            },
        )

        with self.assertRaises(StopAsyncIteration):
            cve = await anext(it)

    async def test_cves_last_modified_end_date(self):
        self.http_client.get.side_effect = create_cpes_responses()

        it = aiter(
            self.api.cpes(
                last_modified_start_date=datetime(2022, 12, 1),
                last_modified_end_date=datetime(2022, 12, 31),
            )
        )
        cve = await anext(it)

        self.assertEqual(cve.cpe_name_id, "CPE-1")
        self.http_client.get.assert_awaited_once_with(
            "https://services.nvd.nist.gov/rest/json/cpes/2.0",
            headers={},
            params={
                "startIndex": 0,
                "lastModStartDate": "2022-12-01T00:00:00",
                "lastModEndDate": "2022-12-31T00:00:00",
            },
        )

        self.http_client.get.reset_mock()

        cve = await anext(it)

        self.assertEqual(cve.cpe_name_id, "CPE-2")
        self.http_client.get.assert_awaited_once_with(
            "https://services.nvd.nist.gov/rest/json/cpes/2.0",
            headers={},
            params={
                "startIndex": 1,
                "lastModStartDate": "2022-12-01T00:00:00",
                "lastModEndDate": "2022-12-31T00:00:00",
                "resultsPerPage": 1,
            },
        )

        with self.assertRaises(StopAsyncIteration):
            cve = await anext(it)

    async def test_cpes_keywords(self):
        self.http_client.get.side_effect = create_cpes_responses()

        it = aiter(self.api.cpes(keywords=["Mac OS X", "kernel"]))
        cve = await anext(it)

        self.assertEqual(cve.cpe_name_id, "CPE-1")
        self.http_client.get.assert_awaited_once_with(
            "https://services.nvd.nist.gov/rest/json/cpes/2.0",
            headers={},
            params={
                "startIndex": 0,
                "keywordSearch": "Mac OS X kernel",
                "keywordExactMatch": "",
            },
        )

        self.http_client.get.reset_mock()

        cve = await anext(it)

        self.assertEqual(cve.cpe_name_id, "CPE-2")
        self.http_client.get.assert_awaited_once_with(
            "https://services.nvd.nist.gov/rest/json/cpes/2.0",
            headers={},
            params={
                "startIndex": 1,
                "resultsPerPage": 1,
                "keywordSearch": "Mac OS X kernel",
                "keywordExactMatch": "",
            },
        )

        with self.assertRaises(StopAsyncIteration):
            cve = await anext(it)

    async def test_cpes_keyword(self):
        self.http_client.get.side_effect = create_cpes_responses()

        it = aiter(self.api.cpes(keywords="macOS"))
        cve = await anext(it)

        self.assertEqual(cve.cpe_name_id, "CPE-1")
        self.http_client.get.assert_awaited_once_with(
            "https://services.nvd.nist.gov/rest/json/cpes/2.0",
            headers={},
            params={
                "startIndex": 0,
                "keywordSearch": "macOS",
            },
        )

        self.http_client.get.reset_mock()

        cve = await anext(it)

        self.assertEqual(cve.cpe_name_id, "CPE-2")
        self.http_client.get.assert_awaited_once_with(
            "https://services.nvd.nist.gov/rest/json/cpes/2.0",
            headers={},
            params={
                "startIndex": 1,
                "resultsPerPage": 1,
                "keywordSearch": "macOS",
            },
        )

        with self.assertRaises(StopAsyncIteration):
            cve = await anext(it)

    async def test_cpes_cpe_match_string(self):
        self.http_client.get.side_effect = create_cpes_responses()

        it = aiter(
            self.api.cpes(
                cpe_match_string="cpe:2.3:o:microsoft:windows_10:20h2:*:*:*:*:*:*:*"
            )
        )
        cve = await anext(it)

        self.assertEqual(cve.cpe_name_id, "CPE-1")
        self.http_client.get.assert_awaited_once_with(
            "https://services.nvd.nist.gov/rest/json/cpes/2.0",
            headers={},
            params={
                "startIndex": 0,
                "cpeMatchString": "cpe:2.3:o:microsoft:windows_10:20h2:*:*:*:*:*:*:*",
            },
        )

        self.http_client.get.reset_mock()

        cve = await anext(it)

        self.assertEqual(cve.cpe_name_id, "CPE-2")
        self.http_client.get.assert_awaited_once_with(
            "https://services.nvd.nist.gov/rest/json/cpes/2.0",
            headers={},
            params={
                "startIndex": 1,
                "resultsPerPage": 1,
                "cpeMatchString": "cpe:2.3:o:microsoft:windows_10:20h2:*:*:*:*:*:*:*",
            },
        )

        with self.assertRaises(StopAsyncIteration):
            cve = await anext(it)

    async def test_cpes_match_criteria_id(self):
        self.http_client.get.side_effect = create_cpes_responses()

        it = aiter(
            self.api.cpes(
                match_criteria_id="36FBCF0F-8CEE-474C-8A04-5075AF53FAF4"
            )
        )
        cve = await anext(it)

        self.assertEqual(cve.cpe_name_id, "CPE-1")
        self.http_client.get.assert_awaited_once_with(
            "https://services.nvd.nist.gov/rest/json/cpes/2.0",
            headers={},
            params={
                "startIndex": 0,
                "matchCriteriaId": "36FBCF0F-8CEE-474C-8A04-5075AF53FAF4",
            },
        )

        self.http_client.get.reset_mock()

        cve = await anext(it)

        self.assertEqual(cve.cpe_name_id, "CPE-2")
        self.http_client.get.assert_awaited_once_with(
            "https://services.nvd.nist.gov/rest/json/cpes/2.0",
            headers={},
            params={
                "startIndex": 1,
                "resultsPerPage": 1,
                "matchCriteriaId": "36FBCF0F-8CEE-474C-8A04-5075AF53FAF4",
            },
        )

        with self.assertRaises(StopAsyncIteration):
            cve = await anext(it)

    async def test_context_manager(self):
        async with self.api:
            pass

        self.http_client.__aenter__.assert_awaited_once()
        self.http_client.__aexit__.assert_awaited_once()



from pathlib import Path
from typing import Iterator, Literal, Tuple, Union

import tomlkit

from ..errors import VersionError
from ..version import Version, VersionUpdate
from ._command import VersionCommand


class CargoVersionCommand(VersionCommand):
    project_file_name = "Cargo.toml"

    def __as_project_document(
        self, origin: Path
    ) -> Iterator[Tuple[Path, tomlkit.TOMLDocument],]:
        """
        Parse the given origin and yields a tuple of path to a
        cargo toml that contains a version

        If the origin is invalid toml than it will raise a VersionError.
        """
        content = origin.read_text(encoding="utf-8")
        content = tomlkit.parse(content)
        package = content.get("package")
        if package:
            version = package.get("version")
            if version:
                yield (origin, content)
        else:
            workspace = content.get("workspace")
            if workspace:
                members = workspace.get("members")
                for member in members:
                    yield from self.__as_project_document(
                        origin.parent / member / self.project_file_name
                    )
        return None

    def update_version(
        self, new_version: Version, *, force: bool = False
    ) -> VersionUpdate:
        try:
            previous_version = self.get_current_version()

            if not force and new_version == previous_version:
                return VersionUpdate(previous=previous_version, new=new_version)
        except VersionError:
            # just ignore current version and override it
            previous_version = None

        changed_files = []
        for project_path, project in self.__as_project_document(
            self.project_file_path
        ):
            project["package"]["version"] = str(new_version)  # type: ignore[index] # noqa: E501
            project_path.write_text(tomlkit.dumps(project))
            changed_files.append(project_path)
        return VersionUpdate(
            previous=previous_version,
            new=new_version,
            changed_files=changed_files,
        )

    def get_current_version(self) -> Version:
        (_, document) = next(self.__as_project_document(self.project_file_path))
        current_version = self.versioning_scheme.parse_version(
            document["package"]["version"]  # type: ignore[index, arg-type]
        )
        return self.versioning_scheme.from_version(current_version)

    def verify_version(
        self, version: Union[Literal["current"], Version, None]
    ) -> None:
        current_version = self.get_current_version()

        if not version or version == "current":
            return

        if current_version != version:
            raise VersionError(
                f"Provided version {version} does not match the "
                f"current version {current_version}."
            )



# Copyright (C) 2022 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# pylint: disable=line-too-long, invalid-name

import unittest

from pontos.github.models.search import (
    InDescriptionQualifier,
    InNameQualifier,
    InReadmeQualifier,
    InTopicsQualifier,
    IsPrivateQualifier,
    IsPublicQualifier,
    NotQualifier,
    OrganizationQualifier,
    RepositoryQualifier,
    UserQualifier,
)


class QualifierTestCase(unittest.TestCase):
    def test_name(self):
        q = InNameQualifier()
        self.assertEqual(str(q), "in:name")

    def test_description(self):
        q = InDescriptionQualifier()
        self.assertEqual(str(q), "in:description")

    def test_topics(self):
        q = InTopicsQualifier()
        self.assertEqual(str(q), "in:topics")

    def test_readme(self):
        q = InReadmeQualifier()
        self.assertEqual(str(q), "in:readme")

    def test_not(self):
        q = NotQualifier(InNameQualifier())
        self.assertEqual(str(q), "-in:name")

    def test_repository(self):
        q = RepositoryQualifier("foo/bar")
        self.assertEqual(str(q), "repo:foo/bar")

    def test_organization(self):
        q = OrganizationQualifier("foo")
        self.assertEqual(str(q), "org:foo")

    def test_user(self):
        q = UserQualifier("foo")
        self.assertEqual(str(q), "user:foo")

    def test_is_public(self):
        q = IsPublicQualifier()
        self.assertEqual(str(q), "is:public")

    def test_is_private(self):
        q = IsPrivateQualifier()
        self.assertEqual(str(q), "is:private")



# SPDX-FileCopyrightText: 2023 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later

import asyncio
from abc import ABC, abstractmethod
from typing import Any, SupportsInt

from pontos.terminal import Terminal


class Command(ABC):
    """Base class for release related command"""

    def __init__(self, *, terminal: Terminal, error_terminal: Terminal) -> None:
        self.terminal = terminal
        self.error_terminal = error_terminal

    def print_error(self, *messages: Any, **kwargs: Any) -> None:
        """Print an error to the error terminal"""
        self.error_terminal.error(*messages, **kwargs)

    def print_warning(self, *messages: Any, **kwargs: Any) -> None:
        """Print a warning to the error console"""
        self.error_terminal.warning(*messages, **kwargs)

    @abstractmethod
    def run(self, **kwargs: Any) -> SupportsInt:
        """Run the command"""


class AsyncCommand(Command):
    """Base class for release related commands using asyncio"""

    def run(self, **kwargs: Any) -> SupportsInt:
        """
        Run the command using asyncio

        MUST NOT be called when an asyncio event loop is already running.
        """
        return asyncio.run(self.async_run(**kwargs))

    @abstractmethod
    async def async_run(self, **kwargs: Any) -> SupportsInt:
        """
        Run the async command

        Gets called via run. Alternatively use similar code as the example.

        Example:
            .. code-block:: python

                import asyncio

                async def main():
                    cmd = MyAsyncCommand()
                    task = asyncio.create_task(cmd.async_run(arg1, arg2))
                    ...
                    await task


                asyncio.run(main())
        """



# SPDX-FileCopyrightText: 2023 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later


import unittest
from pathlib import Path
from unittest.mock import MagicMock, call, patch

from pontos.git import Git
from pontos.release.helper import ReleaseType
from pontos.release.show import (
    OutputFormat,
    ShowReleaseCommand,
    ShowReleaseReturnValue,
)
from pontos.testing import temp_file, temp_git_repository
from pontos.version.schemes import PEP440VersioningScheme


def setup_git_repo(temp_git: Path) -> None:
    some_file = temp_git / "some-file.txt"
    some_file.touch()

    git = Git()
    git.add(some_file)
    git.commit("Add some file", gpg_sign=False, verify=False)
    git.tag("v1.0.0", sign=False)


class ShowTestCase(unittest.TestCase):
    def test_env_output(self):
        terminal = MagicMock()
        with temp_git_repository() as temp_git:
            setup_git_repo(temp_git)

            show_cmd = ShowReleaseCommand(
                terminal=terminal, error_terminal=MagicMock()
            )

            return_val = show_cmd.run(
                versioning_scheme=PEP440VersioningScheme,
                release_type=ReleaseType.PATCH,
                release_version=None,
                git_tag_prefix="v",
                output_format=OutputFormat.ENV,
            )

            self.assertEqual(return_val, ShowReleaseReturnValue.SUCCESS)

            terminal.print.assert_has_calls(
                [
                    call("LAST_RELEASE_VERSION=1.0.0"),
                    call("LAST_RELEASE_VERSION_MAJOR=1"),
                    call("LAST_RELEASE_VERSION_MINOR=0"),
                    call("LAST_RELEASE_VERSION_PATCH=0"),
                    call("RELEASE_VERSION=1.0.1"),
                    call("RELEASE_VERSION_MAJOR=1"),
                    call("RELEASE_VERSION_MINOR=0"),
                    call("RELEASE_VERSION_PATCH=1"),
                ]
            )

            return_val = show_cmd.run(
                versioning_scheme=PEP440VersioningScheme,
                release_type=ReleaseType.MINOR,
                release_version=None,
                git_tag_prefix="v",
                output_format=OutputFormat.ENV,
            )

            self.assertEqual(return_val, ShowReleaseReturnValue.SUCCESS)

            terminal.print.assert_has_calls(
                [
                    call("LAST_RELEASE_VERSION=1.0.0"),
                    call("LAST_RELEASE_VERSION_MAJOR=1"),
                    call("LAST_RELEASE_VERSION_MINOR=0"),
                    call("LAST_RELEASE_VERSION_PATCH=0"),
                    call("RELEASE_VERSION=1.1.0"),
                    call("RELEASE_VERSION_MAJOR=1"),
                    call("RELEASE_VERSION_MINOR=1"),
                    call("RELEASE_VERSION_PATCH=0"),
                ]
            )

            return_val = show_cmd.run(
                versioning_scheme=PEP440VersioningScheme,
                release_type=ReleaseType.MAJOR,
                release_version=None,
                git_tag_prefix="v",
                output_format=OutputFormat.ENV,
            )

            self.assertEqual(return_val, ShowReleaseReturnValue.SUCCESS)

            terminal.print.assert_has_calls(
                [
                    call("LAST_RELEASE_VERSION=1.0.0"),
                    call("LAST_RELEASE_VERSION_MAJOR=1"),
                    call("LAST_RELEASE_VERSION_MINOR=0"),
                    call("LAST_RELEASE_VERSION_PATCH=0"),
                    call("RELEASE_VERSION=2.0.0"),
                    call("RELEASE_VERSION_MAJOR=2"),
                    call("RELEASE_VERSION_MINOR=0"),
                    call("RELEASE_VERSION_PATCH=0"),
                ]
            )

    def test_json_output(self):
        terminal = MagicMock()
        with temp_git_repository() as temp_git:
            setup_git_repo(temp_git)

            show_cmd = ShowReleaseCommand(
                terminal=terminal, error_terminal=MagicMock()
            )

            return_val = show_cmd.run(
                versioning_scheme=PEP440VersioningScheme,
                release_type=ReleaseType.PATCH,
                release_version=None,
                git_tag_prefix="v",
                output_format=OutputFormat.JSON,
            )

            expected = """{
  "release_version": "1.0.1",
  "release_version_major": 1,
  "release_version_minor": 0,
  "release_version_patch": 1,
  "last_release_version": "1.0.0",
  "last_release_version_major": 1,
  "last_release_version_minor": 0,
  "last_release_version_patch": 0
}"""

            self.assertEqual(return_val, ShowReleaseReturnValue.SUCCESS)

            terminal.print.assert_called_once_with(expected)
            terminal.reset_mock()

            return_val = show_cmd.run(
                versioning_scheme=PEP440VersioningScheme,
                release_type=ReleaseType.MINOR,
                release_version=None,
                git_tag_prefix="v",
                output_format=OutputFormat.JSON,
            )

            expected = """{
  "release_version": "1.1.0",
  "release_version_major": 1,
  "release_version_minor": 1,
  "release_version_patch": 0,
  "last_release_version": "1.0.0",
  "last_release_version_major": 1,
  "last_release_version_minor": 0,
  "last_release_version_patch": 0
}"""

            self.assertEqual(return_val, ShowReleaseReturnValue.SUCCESS)

            terminal.print.assert_called_once_with(expected)
            terminal.reset_mock()

            return_val = show_cmd.run(
                versioning_scheme=PEP440VersioningScheme,
                release_type=ReleaseType.MAJOR,
                release_version=None,
                git_tag_prefix="v",
                output_format=OutputFormat.JSON,
            )

            expected = """{
  "release_version": "2.0.0",
  "release_version_major": 2,
  "release_version_minor": 0,
  "release_version_patch": 0,
  "last_release_version": "1.0.0",
  "last_release_version_major": 1,
  "last_release_version_minor": 0,
  "last_release_version_patch": 0
}"""

            self.assertEqual(return_val, ShowReleaseReturnValue.SUCCESS)

            terminal.print.assert_called_once_with(expected)

    def test_github_action_output(self):
        terminal = MagicMock()
        with temp_git_repository() as temp_git:
            setup_git_repo(temp_git)

            show_cmd = ShowReleaseCommand(
                terminal=terminal, error_terminal=MagicMock()
            )

            output_file = temp_git.absolute() / "output"

            with patch.dict(
                "os.environ", {"GITHUB_OUTPUT": f"{output_file}"}, clear=True
            ):
                return_val = show_cmd.run(
                    versioning_scheme=PEP440VersioningScheme,
                    release_type=ReleaseType.PATCH,
                    release_version=None,
                    git_tag_prefix="v",
                    output_format=OutputFormat.GITHUB_ACTION,
                )

                self.assertEqual(return_val, ShowReleaseReturnValue.SUCCESS)

                expected = """last_release_version=1.0.0
last_release_version_major=1
last_release_version_minor=0
last_release_version_patch=0
release_version_major=1
release_version_minor=0
release_version_patch=1
release_version=1.0.1
"""
                actual = output_file.read_text(encoding="utf8")

                self.assertEqual(expected, actual)

                output_file.unlink()

                return_val = show_cmd.run(
                    versioning_scheme=PEP440VersioningScheme,
                    release_type=ReleaseType.MINOR,
                    release_version=None,
                    git_tag_prefix="v",
                    output_format=OutputFormat.GITHUB_ACTION,
                )

                self.assertEqual(return_val, ShowReleaseReturnValue.SUCCESS)

                expected = """last_release_version=1.0.0
last_release_version_major=1
last_release_version_minor=0
last_release_version_patch=0
release_version_major=1
release_version_minor=1
release_version_patch=0
release_version=1.1.0
"""
                actual = output_file.read_text(encoding="utf8")

                self.assertEqual(expected, actual)

                output_file.unlink()

                return_val = show_cmd.run(
                    versioning_scheme=PEP440VersioningScheme,
                    release_type=ReleaseType.MAJOR,
                    release_version=None,
                    git_tag_prefix="v",
                    output_format=OutputFormat.GITHUB_ACTION,
                )

                self.assertEqual(return_val, ShowReleaseReturnValue.SUCCESS)

                expected = """last_release_version=1.0.0
last_release_version_major=1
last_release_version_minor=0
last_release_version_patch=0
release_version_major=2
release_version_minor=0
release_version_patch=0
release_version=2.0.0
"""
                actual = output_file.read_text(encoding="utf8")

                self.assertEqual(expected, actual)

                output_file.unlink()

    def test_initial_release(self):
        terminal = MagicMock()
        with temp_git_repository():
            show_cmd = ShowReleaseCommand(
                terminal=terminal, error_terminal=MagicMock()
            )

            return_val = show_cmd.run(
                versioning_scheme=PEP440VersioningScheme,
                release_type=ReleaseType.VERSION,
                release_version=PEP440VersioningScheme.parse_version("1.0.0"),
                git_tag_prefix="v",
                output_format=OutputFormat.ENV,
            )

            self.assertEqual(return_val, ShowReleaseReturnValue.SUCCESS)

            terminal.print.assert_has_calls(
                [
                    call("LAST_RELEASE_VERSION="),
                    call("LAST_RELEASE_VERSION_MAJOR="),
                    call("LAST_RELEASE_VERSION_MINOR="),
                    call("LAST_RELEASE_VERSION_PATCH="),
                    call("RELEASE_VERSION=1.0.0"),
                    call("RELEASE_VERSION_MAJOR=1"),
                    call("RELEASE_VERSION_MINOR=0"),
                    call("RELEASE_VERSION_PATCH=0"),
                ]
            )

            terminal.reset_mock()

            return_val = show_cmd.run(
                versioning_scheme=PEP440VersioningScheme,
                release_type=ReleaseType.VERSION,
                release_version=PEP440VersioningScheme.parse_version("1.0.0"),
                git_tag_prefix="v",
                output_format=OutputFormat.JSON,
            )

            expected = """{
  "release_version": "1.0.0",
  "release_version_major": 1,
  "release_version_minor": 0,
  "release_version_patch": 0,
  "last_release_version": "",
  "last_release_version_major": "",
  "last_release_version_minor": "",
  "last_release_version_patch": ""
}"""

            self.assertEqual(return_val, ShowReleaseReturnValue.SUCCESS)

            terminal.print.assert_called_once_with(expected)

            with temp_file(name="output") as output_file, patch.dict(
                "os.environ", {"GITHUB_OUTPUT": f"{output_file}"}, clear=True
            ):
                return_val = show_cmd.run(
                    versioning_scheme=PEP440VersioningScheme,
                    release_type=ReleaseType.VERSION,
                    release_version=PEP440VersioningScheme.parse_version(
                        "1.0.0"
                    ),
                    git_tag_prefix="v",
                    output_format=OutputFormat.GITHUB_ACTION,
                )

                self.assertEqual(return_val, ShowReleaseReturnValue.SUCCESS)

                expected = """last_release_version=
last_release_version_major=
last_release_version_minor=
last_release_version_patch=
release_version_major=1
release_version_minor=0
release_version_patch=0
release_version=1.0.0
"""
                actual = output_file.read_text(encoding="utf8")

                self.assertEqual(expected, actual)



# Copyright (C) 2022 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from typing import Optional

from pontos.github.api.client import GitHubAsyncREST


class GitHubAsyncRESTContent(GitHubAsyncREST):
    async def path_exists(
        self, repo: str, path: str, *, branch: Optional[str] = None
    ) -> bool:
        """
        Check if a path (file or directory) exists in a branch of a repository

        Args:
            repo: GitHub repository (owner/name) to use
            path: to the file/directory in question
            branch: Branch to check, defaults to default branch (:

        Returns:
            True if existing, False else

        Example:
            .. code-block:: python

                from pontos.github.api import GitHubAsyncRESTApi

                async with GitHubAsyncRESTApi(token) as api:
                    exists = await api.contents.path_exists(
                        "foo/bar", "src/utils.py"
                    )
        """
        api = f"/repos/{repo}/contents/{path}"
        params = {}
        if branch:
            params["ref"] = branch

        response = await self._client.get(api, params=params)
        return response.is_success
